from __future__ import annotations

from typing import Optional

from psnawp_api.models.trophies.trophy_constants import TrophyRarity, TrophyType


def trophy_type_str_to_enum(trophy_type_str: Optional[str]) -> Optional[TrophyType]:
    return TrophyType(trophy_type_str) if trophy_type_str is not None else None


def trophy_rarity_to_enum(trophy_rarity: Optional[int]) -> Optional[TrophyRarity]:
    return TrophyRarity(trophy_rarity) if trophy_rarity is not None else None



from psnawp_api.utils.endpoints import API_PATH, BASE_PATH
from psnawp_api.utils.misc import extract_region_from_npid, iso_format_to_datetime

__all__ = ["BASE_PATH", "API_PATH", "iso_format_to_datetime", "extract_region_from_npid"]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar

from attrs import define, field
from typing_extensions import Self

from psnawp_api.core import PSNAWPNotFound
from psnawp_api.models.trophies import PlatformType, TrophySet
from psnawp_api.utils import API_PATH, BASE_PATH, iso_format_to_datetime

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator


@define(frozen=True)
class TrophyGroupSummary:
    """TrophyGroupSummary contains trophy count data for one trophy group of a game title."""

    trophy_group_id: Optional[str]
    "ID for the trophy group (all titles have default, additional groups are 001 incrementing)"
    trophy_group_name: Optional[str]
    "Trophy group name"
    trophy_group_detail: Optional[str]
    "Trophy group description (PS3, PS4 and PS Vita titles only)"
    trophy_group_icon_url: Optional[str]
    "URL of the icon for the trophy group"
    defined_trophies: TrophySet
    "Number of trophies for the trophy group by type"

    @classmethod
    def from_dict(cls, trophy_group_dict: dict[str, Any]) -> Self:
        return cls(
            trophy_group_id=trophy_group_dict.get("trophyGroupId"),
            trophy_group_name=trophy_group_dict.get("trophyGroupName"),
            trophy_group_detail=trophy_group_dict.get("trophyGroupDetail"),
            trophy_group_icon_url=trophy_group_dict.get("trophyGroupIconUrl"),
            defined_trophies=TrophySet(
                **trophy_group_dict.get(
                    "definedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
        )


@define(frozen=True)
class TrophyGroupSummaryWithProgress(TrophyGroupSummary):
    """TrophyGroupSummaryWithProgress contains trophy count data for one trophy group of a game title and user progress for each trophy group."""

    progress: Optional[int]
    "Percentage of trophies earned for group"
    earned_trophies: TrophySet
    "Number of trophies for the group which have been earned by type"
    last_updated_datetime: Optional[datetime] = field(converter=iso_format_to_datetime)
    "Date most recent trophy earned for the group"

    @classmethod
    def from_dict(cls, trophy_group_dict: dict[str, Any]) -> Self:
        return cls(
            trophy_group_id=trophy_group_dict.get("trophyGroupId"),
            trophy_group_name=trophy_group_dict.get("trophyGroupName"),
            trophy_group_detail=trophy_group_dict.get("trophyGroupDetail"),
            trophy_group_icon_url=trophy_group_dict.get("trophyGroupIconUrl"),
            defined_trophies=TrophySet(
                **trophy_group_dict.get(
                    "definedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
            progress=trophy_group_dict.get("progress"),
            earned_trophies=TrophySet(
                **trophy_group_dict.get(
                    "earnedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
            last_updated_datetime=trophy_group_dict.get("lastUpdatedDateTime"),
        )


T = TypeVar("T", bound=TrophyGroupSummary)


class TrophyGroupsSummary(Generic[T]):
    """TrophyGroupsSummary contains trophy count data for all the groups in a game title combined and individual."""

    def __init__(
        self,
        trophy_set_version: Optional[str],
        trophy_title_name: Optional[str],
        trophy_title_detail: Optional[str],
        trophy_title_icon_url: Optional[str],
        trophy_title_platform: set[PlatformType],
        defined_trophies: TrophySet,
        trophy_groups: list[T],
        hidden_flag: Optional[bool],
        progress: Optional[int],
        earned_trophies: TrophySet,
        last_updated_date_time: Optional[str],
    ) -> None:
        self.trophy_set_version = trophy_set_version
        "The current version of the trophy set"
        self.trophy_title_name = trophy_title_name
        "Title name"
        self.trophy_title_detail = trophy_title_detail
        "Title description (PS3, PS4 and PS Vita titles only)"
        self.trophy_title_icon_url = trophy_title_icon_url
        "URL of the icon for the trophy title"
        self.trophy_title_platform = trophy_title_platform
        "The platform this title belongs to"
        self.defined_trophies = defined_trophies
        "Total number of trophies for the title by type"
        self.trophy_groups = trophy_groups
        "Individual object for each trophy group returned"
        self.hidden_flag = hidden_flag
        "Whether title has been hidden on the accounts trophy list (Authenticating account only)"
        self.progress = progress
        "Percentage of trophies earned for the title"
        self.earned_trophies = earned_trophies
        "Number of trophies for the title which have been earned by type"
        self.last_updated_date_time = iso_format_to_datetime(last_updated_date_time)
        "Date most recent trophy earned for the title (UTC+00:00 TimeZone)"

    def __str__(self) -> str:
        return (
            f"TrophyGroupsSummary(Title: {self.trophy_title_name}, Defined: {self.defined_trophies}, Earned: {self.earned_trophies}, Progress: {self.progress})"
        )

    def __repr__(self) -> str:
        return (
            f"TrophyGroupsSummary(trophy_set_version={self.trophy_set_version}, "
            f"trophy_title_name={self.trophy_title_name}, "
            f"trophy_title_detail={self.trophy_title_detail}, "
            f"trophy_title_icon_url={self.trophy_title_icon_url}, "
            f"trophy_title_platform={self.trophy_title_platform}, "
            f"defined_trophies={self.defined_trophies}, "
            f"trophy_groups={self.trophy_groups}, "
            f"hidden_flag={self.hidden_flag}, "
            f"progress={self.progress}, "
            f"earned_trophies={self.earned_trophies}, "
            f"last_updated_date_time={self.last_updated_date_time})"
        )


class TrophyGroupsSummaryBuilder:
    """Class for providing convenient method to Build TrophyGroupsSummary from PlayStation Endpoints"""

    def __init__(self, authenticator: Authenticator, np_communication_id: str):
        """:param request_builder: The instance of authenticator. Used to make HTTPRequests.
        :param np_communication_id: Unique ID of a game title used to request trophy information. This can be obtained from ``GameTitle`` class.

        """
        self.authenticator = authenticator
        self.np_communication_id: str = np_communication_id

    @staticmethod
    def trophy_groups_dict_to_obj(trophy_groups_dict: dict[str, Any], trophy_groups: list[T]) -> TrophyGroupsSummary[T]:
        """Takes list of TrophyGroupSummary and TrophyGroupSummaryWithProgress and returns TrophyGroupsSummary[TrophyGroupSummary]
            or TrophyGroupsSummary[TrophyGroupSummaryWithProgress].

        :param trophy_groups_dict: dict from endpoint.
        :param trophy_groups: list of TrophyGroupSummary or TrophyGroupSummaryWithProgress

        :returns: TrophyGroupsSummary containing list of TrophyGroupSummary or TrophyGroupSummaryWithProgress along with some more information.

        """
        trophy_group_summary = TrophyGroupsSummary(
            trophy_set_version=trophy_groups_dict.get("trophySetVersion"),
            trophy_title_name=trophy_groups_dict.get("trophyTitleName"),
            trophy_title_detail=trophy_groups_dict.get("trophyTitleDetail"),
            trophy_title_icon_url=trophy_groups_dict.get("trophyTitleIconUrl"),
            trophy_title_platform=set(
                [PlatformType(platform_str) for platform_str in trophy_groups_dict.get("trophyTitlePlatform", "").split(",")],
            ),
            defined_trophies=TrophySet(
                **trophy_groups_dict.get(
                    "definedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
            trophy_groups=trophy_groups,
            hidden_flag=trophy_groups_dict.get("hiddenFlag"),
            progress=trophy_groups_dict.get("progress"),
            earned_trophies=TrophySet(
                **trophy_groups_dict.get(
                    "earnedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
            last_updated_date_time=trophy_groups_dict.get("lastUpdatedDateTime"),
        )
        return trophy_group_summary

    def game_title_trophy_groups_summary(
        self,
        platform: PlatformType,
    ) -> TrophyGroupsSummary[TrophyGroupSummary]:
        """Retrieves the trophy groups for a title and their respective trophy count.

        This is most commonly seen in games which have expansions where additional trophies are added.

        :param platform: The platform this title belongs to.

        :returns: TrophyGroupSummary object containing title and title groups trophy information.

        :raises PSNAWPNotFound: if you don't have any trophies for that game.
        :raises PSNAWPForbidden: If the user's profile is private.

        """

        service_name = "trophy2" if platform == PlatformType.PS5 else "trophy"
        params = {"npServiceName": service_name}
        try:
            response = self.authenticator.get(
                url=f"{BASE_PATH['trophies']}{API_PATH['title_trophy_group'].format(np_communication_id=self.np_communication_id)}",
                params=params,
            ).json()
        except PSNAWPNotFound as not_found:
            raise PSNAWPNotFound("The following user has no trophies for the given game title.") from not_found

        trophy_groups = [TrophyGroupSummary.from_dict(trophy_group) for trophy_group in response.get("trophyGroups", [])]

        return type(self).trophy_groups_dict_to_obj(
            response,
            trophy_groups,
        )

    def earned_user_trophy_groups_summary(
        self,
        account_id: str,
        platform: PlatformType,
    ) -> TrophyGroupsSummary[TrophyGroupSummaryWithProgress]:
        """Retrieves the earned trophy groups for a title and their respective trophy count along with their trophy earned progress.

        This is most commonly seen in games which have expansions where additional trophies are added.

        :param account_id: The account whose trophy list is being accessed.
        :param platform: The platform this title belongs to.

        :returns: TrophyGroupSummary object containing title and title groups trophy information.

        :raises PSNAWPNotFound: if you don't have any trophies for that game.
        :raises PSNAWPForbidden: If the user's profile is private.

        """

        service_name = "trophy2" if platform == PlatformType.PS5 else "trophy"
        params = {"npServiceName": service_name}

        trophy_groups_metadata = self.authenticator.get(
            url=f"{BASE_PATH['trophies']}{API_PATH['title_trophy_group'].format(np_communication_id=self.np_communication_id)}",
            params=params,
        ).json()

        trophy_groups_user_data = self.authenticator.get(
            url=f"{BASE_PATH['trophies']}{API_PATH['user_title_trophy_group'].format(account_id=account_id, np_communication_id=self.np_communication_id)}",
            params=params,
        ).json()

        merged_data: dict[str, Any] = trophy_groups_metadata | trophy_groups_user_data
        merged_trophy_groups: list[dict[str, Any]] = [
            x[0] | x[1] for x in zip(trophy_groups_metadata.get("trophyGroups"), trophy_groups_user_data.get("trophyGroups"))
        ]
        merged_data["trophyGroups"] = merged_trophy_groups

        trophy_groups = [TrophyGroupSummaryWithProgress.from_dict(trophy_group) for trophy_group in merged_data.get("trophyGroups", [])]

        return type(self).trophy_groups_dict_to_obj(
            merged_data,
            trophy_groups,
        )



from __future__ import annotations

from typing import Any, Generator, Iterable, Optional, overload

from psnawp_api.core import Authenticator, PSNAWPIllegalArgumentError, RequestBuilderHeaders
from psnawp_api.models import Client, GameTitle, Group, SearchDomain, UniversalSearch, User
from psnawp_api.models.listing import PaginationArguments
from psnawp_api.models.search import SearchResult


class PSNAWP:
    """PlayStation Network API Wrapper Python (PSNAWP) Retrieve User Information, Trophies, Game and Store data from the PlayStation Network.

    Instances of this class are the gateway to interacting with PSN API through PSNAWP.

    .. code-block:: Python

        from psnawp_api import PSNAWP

        psnawp = PSNAWP("<64 character npsso code>")

    """

    def __init__(
        self,
        npsso_cookie: str,
        headers: Optional[RequestBuilderHeaders] = None,
    ) -> None:
        """Constructor Method. Takes the npsso_cookie and creates instance of ``request_builder.RequestBuilder`` which is used later in code for HTTPS requests.

        :param npsso_cookie: npsso cookie obtained from PSN website.
        :param headers: Common headers that will be added to all HTTP request such as ``Country`` and ``Accept-Language``.

        :raises PSNAWPAuthenticationError: If npsso code is expired or is incorrect.

        """

        default_headers: RequestBuilderHeaders = {
            "User-Agent": "Mozilla/5.0 (Linux; Android 11; sdk_gphone_x86 Build/RSR1.201013.001; wv) \
            AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/83.0.4103.106 Mobile Safari/537.36",
            "Accept-Language": "en-US,en;q=0.9",
            "Country": "US",
        }

        _header = default_headers | headers if headers is not None else default_headers
        self.authenticator = Authenticator(
            npsso_cookie=npsso_cookie,
            common_headers=_header,
        )

    def me(self) -> Client:
        """Creates a new client object (your account).

        :returns: Client Object

        .. code-block:: Python

            from psnawp_api import PSNAWP

            psnawp = PSNAWP("<64 character npsso code>")
            client = psnawp.me()

        """
        return Client(self.authenticator)

    @overload
    def user(self, *, online_id: str) -> User: ...

    @overload
    def user(self, *, account_id: str) -> User: ...

    def user(self, **kwargs: Any) -> User:
        """Creates a new user object using Online ID (GamerTag) or Account ID (PSN ID).

        .. note::

            The account_id takes higher precedence than online_id. If both arguments are passed, online_id will be ignored.

        :param online_id: Online ID (GamerTag) of the user.
        :param account_id: Account ID of the user.

        :returns: User Object

        :raises PSNAWPIllegalArgumentError: If None of the kwargs are passed.
        :raises PSNAWPNotFound: If the online_id or account_id is not valid/found.

        .. code-block:: Python

            user1 = psnawp.user(online_id="VaultTec_Trading")
            user2 = psnawp.user(account_id="1802043923080044300")

        """
        online_id: Optional[str] = kwargs.get("online_id")
        account_id: Optional[str] = kwargs.get("account_id")

        if account_id is not None:
            return User.from_account_id(self.authenticator, account_id)
        elif online_id is not None:
            return User.from_online_id(self.authenticator, online_id)
        else:
            raise PSNAWPIllegalArgumentError("You must provide at least online ID or account ID.")

    def game_title(self, title_id: str, account_id: str = "6515971742264256071", np_communication_id: Optional[str] = None) -> GameTitle:
        """Creates a GameTitle class object from title_id which represents a PlayStation video game title.

        .. note::

            The GameTitle class is only useful if the user has played that video game. To allow users to retrieve information without having to play that video
            game I picked a default user who has played the most number of games based on this website
            (https://www.truetrophies.com/leaderboard/gamer/gamesplayed). It is possible that the there are games this user has not played and in that case it
            is better to provide your own account id (``'me'``) or someone who has played that game.

        .. note::

            ``title_id`` can be obtained from https://andshrew.github.io/PlayStation-Titles/ or from :py:class:`psnawp_api.models.search.Search`

        .. note::

            During the construction of the object, an additional call is made to get the np_communication_id. This ID is important for getting trophy data. This
            call can be skipped by providing np_communication_id in as argument.

        :param title_id: unique ID of game title.
        :param: account_id: The account whose trophy list is being accessed
        :param np_communication_id: Unique ID of a game title used to request trophy information.

        :returns: Title Object

        :raises PSNAWPNotFound: If the user does not have any trophies for that game or the game doesn't exist.

        """
        return GameTitle(self.authenticator, title_id=title_id, account_id=account_id, np_communication_id=np_communication_id)

    @overload
    def group(self, *, group_id: str) -> Group: ...

    @overload
    def group(self, *, users_list: Iterable[User]) -> Group: ...

    def group(self, **kwargs: Any) -> Group:
        """Creates a group object from a Group ID or from list of users.

        .. warning::

            Passing ``users_list`` will create a new group each time. If you want to continue from the same group. Use group id obtained from
            ``client.get_groups()``

        :param group_id: The Group ID of a group usually retrieved with the ``get_groups()`` method.
        :param users_list: A list of users of the members in the group.

        :returns: Group Object

        :raises PSNAWPNotFound: If group id does not exist or is invalid.
        :raises PSNAWPForbidden: If you are sending message a user who has blocked you.

        """

        group_id: Optional[str] = kwargs.get("group_id")
        users: Optional[Iterable[User]] = kwargs.get("users_list")

        if group_id is not None:
            return Group.create_from_group_id(self.authenticator, group_id=group_id)
        elif users is not None:
            return Group.create_from_users(self.authenticator, users=users)
        else:
            raise PSNAWPIllegalArgumentError("You provide at least Group Id or Users")

    def search(
        self,
        search_query: str,
        search_domain: SearchDomain,
        limit: Optional[int] = None,
        offset: int = 0,
        page_size: int = 20,
    ) -> Generator[SearchResult, None, None]:
        """Creates a new search object

        :param search_query: _description_
        :param search_domain: _description_
        :param limit: Total numbers of items to receive, None means no limit.
        :param page_size: The number of items to receive per api request.
        :param offset: Specifies the offset for paginator.

        :returns: Search Iterator object to iterate over search results.

        """
        pg_args = PaginationArguments(total_limit=limit, offset=offset, page_size=page_size)
        if search_domain == SearchDomain.FULL_GAMES:
            return UniversalSearch(authenticator=self.authenticator, pagination_args=pg_args, search_query=search_query).search_full_game()
        else:
            return UniversalSearch(authenticator=self.authenticator, pagination_args=pg_args, search_query=search_query).search_add_onns()



from psnawp_api.models.trophies.trophy import Trophy, TrophyIterator, TrophyWithProgress, TrophyWithProgressIterator
from psnawp_api.models.trophies.trophy_constants import PlatformType, TrophySet
from psnawp_api.models.trophies.trophy_group import TrophyGroupsSummary, TrophyGroupsSummaryBuilder, TrophyGroupSummary, TrophyGroupSummaryWithProgress
from psnawp_api.models.trophies.trophy_summary import TrophySummary
from psnawp_api.models.trophies.trophy_titles import TrophyTitle, TrophyTitleIterator

__all__ = [
    "PlatformType",
    "Trophy",
    "TrophyGroupsSummary",
    "TrophyGroupsSummaryBuilder",
    "TrophyGroupSummary",
    "TrophyGroupSummaryWithProgress",
    "TrophyIterator",
    "TrophySet",
    "TrophySummary",
    "TrophyTitle",
    "TrophyTitleIterator",
    "TrophyWithProgress",
    "TrophyWithProgressIterator",
]



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generator, Optional, TypedDict

from typing_extensions import Self

from psnawp_api.models.listing import PaginationIterator
from psnawp_api.models.trophies.trophy_constants import PlatformType, TrophyRarity, TrophyType
from psnawp_api.models.trophies.utility_functions import (
    trophy_rarity_to_enum,
    trophy_type_str_to_enum,
)
from psnawp_api.utils import iso_format_to_datetime
from psnawp_api.utils.endpoints import API_PATH, BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.listing import PaginationArguments


@dataclass(frozen=True)
class Trophy:
    """A class that represents a PlayStation Video Game Trophy."""

    # Trophy Group Metadata
    trophy_set_version: Optional[str]
    "The current version of the trophy set"
    has_trophy_groups: Optional[bool]
    "True if this title has additional trophy groups"

    # Trophy Meta
    trophy_id: Optional[int]
    "Unique ID for this trophy"
    trophy_hidden: Optional[bool]
    "True if this is a secret trophy (Only for client)"
    trophy_type: Optional[TrophyType]
    "Type of the trophy"
    trophy_name: Optional[str]
    "Name of trophy"
    trophy_detail: Optional[str]
    "Description of the trophy"
    trophy_icon_url: Optional[str]
    "URL for the graphic associated with the trophy"
    trophy_group_id: Optional[str]
    "ID of the trophy group this trophy belongs to"
    trophy_progress_target_value: Optional[int]
    "Trophy progress towards it being unlocked (PS5 Only)"
    trophy_reward_name: Optional[str]
    "Name of the reward earning the trophy grants (PS5 Only)"
    trophy_reward_img_url: Optional[str]
    "URL for the graphic associated with the reward (PS5 Only)"

    @classmethod
    def from_trophy_dict(cls, trophy_dict: dict[str, Any]) -> Self:
        trophy_instance = cls(
            trophy_set_version=trophy_dict.get("trophySetVersion"),
            has_trophy_groups=trophy_dict.get("hasTrophyGroups"),
            trophy_id=trophy_dict.get("trophyId"),
            trophy_hidden=trophy_dict.get("trophyHidden"),
            trophy_type=trophy_type_str_to_enum(trophy_dict.get("trophyType")),
            trophy_name=trophy_dict.get("trophyName"),
            trophy_detail=trophy_dict.get("trophyDetail"),
            trophy_icon_url=trophy_dict.get("trophyIconUrl"),
            trophy_group_id=trophy_dict.get("trophyGroupId"),
            trophy_progress_target_value=trophy_dict.get("trophyProgressTargetValue"),
            trophy_reward_name=trophy_dict.get("trophyRewardName"),
            trophy_reward_img_url=trophy_dict.get("trophyRewardImageUrl"),
        )
        return trophy_instance


@dataclass(frozen=True)
class TrophyWithProgress(Trophy):
    # Earned Trophy Info
    earned: Optional[bool]
    "True if this trophy has been earned"
    progress: Optional[int]
    "If the trophy tracks progress towards unlock this is number of steps currently completed (ie. 73/300) (PS5 titles only)"
    progress_rate: Optional[int]
    "If the trophy tracks progress towards unlock this is the current percentage complete (PS5 titles only)"
    progressed_date_time: Optional[datetime]
    "If the trophy tracks progress towards unlock, and some progress has been made, then this returns the date progress was last updated. (PS5 titles only)"
    earned_date_time: Optional[datetime]
    "Date trophy was earned"
    trophy_rarity: Optional[TrophyRarity]
    "Rarity of the trophy"
    trophy_earn_rate: Optional[float]
    "Percentage of all users who have earned the trophy"

    @classmethod
    def from_trophy_dict(cls, trophy_dict: dict[str, Any]) -> Self:
        trophy_instance = cls(
            trophy_set_version=trophy_dict.get("trophySetVersion"),
            has_trophy_groups=trophy_dict.get("hasTrophyGroups"),
            trophy_id=trophy_dict.get("trophyId"),
            trophy_hidden=trophy_dict.get("trophyHidden"),
            trophy_type=trophy_type_str_to_enum(trophy_dict.get("trophyType")),
            trophy_name=trophy_dict.get("trophyName"),
            trophy_detail=trophy_dict.get("trophyDetail"),
            trophy_icon_url=trophy_dict.get("trophyIconUrl"),
            trophy_group_id=trophy_dict.get("trophyGroupId"),
            trophy_progress_target_value=trophy_dict.get("trophyProgressTargetValue"),
            trophy_reward_name=trophy_dict.get("trophyRewardName"),
            trophy_reward_img_url=trophy_dict.get("trophyRewardImageUrl"),
            earned=trophy_dict.get("earned"),
            progress=trophy_dict.get("progress"),
            progress_rate=trophy_dict.get("progressRate"),
            progressed_date_time=iso_format_to_datetime(trophy_dict.get("progressedDateTime")),
            earned_date_time=iso_format_to_datetime(trophy_dict.get("earnedDateTime")),
            trophy_rarity=trophy_rarity_to_enum(trophy_dict.get("trophyRare")),
            trophy_earn_rate=trophy_dict.get("trophyEarnedRate"),
        )
        return trophy_instance

    @classmethod
    def from_trophies_list(cls, trophies_dict: Optional[list[dict[str, Any]]]) -> list[Trophy]:
        trophy_list: list[Trophy] = []
        if trophies_dict is None:
            return trophy_list

        for trophy_dict in trophies_dict:
            trophy_instance = cls.from_trophy_dict(trophy_dict)
            trophy_list.append(trophy_instance)
        return trophy_list


class TrophyIterator(PaginationIterator[Trophy]):
    """Class for Iterating over all trophies for a specified group within a game title."""

    def __init__(
        self,
        authenticator: Authenticator,
        url: str,
        pagination_args: PaginationArguments,
        platform: PlatformType,
    ) -> None:
        super().__init__(authenticator=authenticator, url=url, pagination_args=pagination_args)
        self.platform = platform

    @classmethod
    def from_endpoint(
        cls,
        authenticator: Authenticator,
        pagination_args: PaginationArguments,
        np_communication_id: str,
        platform: PlatformType,
        trophy_group_id: str,
    ) -> Self:
        url = f"{BASE_PATH['trophies']}{API_PATH['trophies_for_title'].format(np_communication_id=np_communication_id, trophy_group_id=trophy_group_id)}"
        return cls(authenticator, url, pagination_args, platform)

    def fetch_next_page(self) -> Generator[Trophy, None, None]:
        service_name = "trophy2" if self.platform == PlatformType.PS5 else "trophy"
        params = {"npServiceName": service_name} | self._pagination_args.get_params_dict()
        response = self.authenticator.get(url=self._url, params=params).json()
        self._total_item_count = response.get("totalItemCount", 0)

        trophies: list[dict[str, Any]] = response.get("trophies", [])
        for trophy in trophies:
            trophy_instance = Trophy.from_trophy_dict(
                {
                    **trophy,
                    "trophySetVersion": response.get("trophySetVersion"),
                    "hasTrophyGroups": response.get("hasTrophyGroups"),
                }
            )
            self._pagination_args.increment_offset()
            yield trophy_instance

        offset = response.get("nextOffset") or 0
        if offset > 0:
            self._has_next = True
        else:
            self._has_next = False


class RarestTrophies(TypedDict):
    trophyId: int
    trophyHidden: bool
    earned: bool
    earnedDateTime: str
    trophyType: str
    trophyRare: int
    trophyEarnedRate: str


class TrophyWithProgressIterator(PaginationIterator[TrophyWithProgress]):
    """Class for Iterating over all trophies for a specified group within a game title, this class includes user progress for each trophy."""

    def __init__(
        self,
        authenticator: Authenticator,
        url: str,
        pagination_args: PaginationArguments,
        platform: PlatformType,
        progress_url: str,
    ) -> None:
        super().__init__(authenticator=authenticator, url=url, pagination_args=pagination_args)
        self.platform = platform
        self._progress_url = progress_url

        self.rarest_trophies: Optional[list[RarestTrophies]] = None

    @classmethod
    def from_endpoint(
        cls,
        authenticator: Authenticator,
        pagination_args: PaginationArguments,
        np_communication_id: str,
        platform: PlatformType,
        trophy_group_id: str,
        account_id: str,
    ) -> Self:
        url = f"{BASE_PATH['trophies']}{API_PATH['trophies_for_title'].format(np_communication_id=np_communication_id, trophy_group_id=trophy_group_id)}"
        progress_url = (
            f"{BASE_PATH['trophies']}"
            f"{API_PATH['trophies_earned_for_title'].format(account_id=account_id, np_communication_id=np_communication_id, trophy_group_id=trophy_group_id)}"
        )
        return cls(authenticator, url, pagination_args, platform, progress_url)

    def fetch_next_page(self) -> Generator[TrophyWithProgress, None, None]:
        service_name = "trophy2" if self.platform == PlatformType.PS5 else "trophy"
        params = {"npServiceName": service_name} | self._pagination_args.get_params_dict()

        response = self.authenticator.get(url=self._url, params=params).json()
        self._total_item_count = response.get("totalItemCount", 0)
        trophies: list[dict[str, Any]] = response.get("trophies")

        response_progress = self.authenticator.get(url=self._progress_url, params=params).json()
        self.rarest_trophies = response_progress.get("rarestTrophies")
        trophies_progress: list[dict[str, Any]] = response_progress.get("trophies")

        for trophy, progress in zip(trophies, trophies_progress):
            trophy_instance = TrophyWithProgress.from_trophy_dict(
                {
                    **trophy,
                    **progress,
                    "trophySetVersion": response.get("trophySetVersion"),
                    "hasTrophyGroups": response.get("hasTrophyGroups"),
                }
            )
            self._pagination_args.increment_offset()
            yield trophy_instance

        offset = response.get("nextOffset") or 0
        if offset > 0:
            self._has_next = True
        else:
            self._has_next = False



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Generator, Optional

from attrs import define, field
from typing_extensions import Self

from psnawp_api.core import PSNAWPBadRequest, PSNAWPNotFound
from psnawp_api.models.listing import PaginationIterator
from psnawp_api.models.trophies import PlatformType, TrophySet
from psnawp_api.utils import API_PATH, BASE_PATH, iso_format_to_datetime

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.listing import PaginationArguments


@define(frozen=True)
class TrophyTitle:
    """A class containing summary of trophy data for a user for a game title"""

    np_service_name: Optional[str]
    "trophy for PS3, PS4, or PS Vita platforms and trophy2 for the PS5 platform"
    np_communication_id: Optional[str]
    "Unique ID of the title"
    trophy_set_version: Optional[str]
    "The current version of the trophy set"
    title_name: Optional[str]
    "Title name"
    title_detail: Optional[str]
    "Title description (PS3, PS4 and PS Vita titles only)"
    title_icon_url: Optional[str]
    "URL of the icon for the title"
    title_platform: frozenset[PlatformType]
    "Platforms this title belongs to"
    has_trophy_groups: Optional[bool]
    "True if the title has multiple groups of trophies (eg. DLC trophies which are separate from the main trophy list)"
    progress: Optional[int]
    "Percentage of trophies earned for the title"
    hidden_flag: Optional[bool]
    "Title has been hidden on the accounts trophy list (Only for Client)"
    earned_trophies: TrophySet
    "Number of trophies for the title which have been earned by type"
    defined_trophies: TrophySet
    "Number of trophies for the title by type"
    last_updated_date_time: Optional[datetime] = field(converter=iso_format_to_datetime)
    "Date most recent trophy earned for the title (UTC+00:00 TimeZone)"
    # when title_id is passed
    np_title_id: Optional[str]
    "Title ID of the title if passed"


class TrophyTitleIterator(PaginationIterator[TrophyTitle]):
    """Retrieve all game titles associated with an account, and a summary of trophies earned from them."""

    def __init__(self, authenticator: Authenticator, url: str, pagination_args: PaginationArguments, title_ids: Optional[list[str]]) -> None:
        """Constructor of TrophyTitles class.

        .. note::

            This class is intended to be interfaced with through PSNAWP.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param url: The url of endpoint.
        :param pagination_args: Arguments related to pagination like limit and offset.

        """
        super().__init__(authenticator=authenticator, url=url, pagination_args=pagination_args)
        self.title_ids = title_ids

    @classmethod
    def from_endpoint(
        cls,
        authenticator: Authenticator,
        pagination_args: PaginationArguments,
        account_id: str,
        title_ids: Optional[list[str]],
    ) -> Self:
        if title_ids is None:
            url = f"{BASE_PATH['trophies']}{API_PATH['trophy_titles'].format(account_id=account_id)}"
        else:
            url = f"{BASE_PATH['trophies']}{API_PATH['trophy_titles_for_title'].format(account_id=account_id)}"
        return cls(
            authenticator,
            url,
            pagination_args,
            title_ids,
        )

    def fetch_next_page(self) -> Generator[TrophyTitle, None, None]:
        if self.title_ids is None:
            return self.get_trophy_title()
        else:
            return self.get_trophy_summary_for_title()

    def get_trophy_title(self) -> Generator[TrophyTitle, None, None]:
        """Retrieve all game titles associated with an account, and a summary of trophies earned from them.

        :param limit: Limit of titles returned, None means to return all trophy titles.

        :returns: Generator object with TitleTrophySummary objects

        :raises PSNAWPForbidden: If the user's profile is private

        """
        response = self.authenticator.get(url=self._url, params=self._pagination_args.get_params_dict()).json()
        self._total_item_count = response.get("totalItemCount", 0)

        trophy_titles: list[dict[str, Any]] = response.get("trophyTitles")
        for trophy_title in trophy_titles:
            title_trophy_sum = TrophyTitle(
                np_service_name=trophy_title.get("npServiceName"),
                np_communication_id=trophy_title.get("npCommunicationId"),
                trophy_set_version=trophy_title.get("trophySetVersion"),
                title_name=trophy_title.get("trophyTitleName"),
                title_detail=trophy_title.get("trophyTitleDetail"),
                title_icon_url=trophy_title.get("trophyTitleIconUrl"),
                title_platform=frozenset(
                    [PlatformType(platform_str) for platform_str in trophy_title.get("trophyTitlePlatform", "").split(",")],
                ),
                has_trophy_groups=trophy_title.get("hasTrophyGroups"),
                progress=trophy_title.get("progress"),
                hidden_flag=trophy_title.get("hiddenFlag"),
                last_updated_date_time=trophy_title.get("lastUpdatedDateTime"),
                defined_trophies=TrophySet(
                    **trophy_title.get(
                        "definedTrophies",
                        {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                    ),
                ),
                earned_trophies=TrophySet(
                    **trophy_title.get(
                        "earnedTrophies",
                        {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                    )
                ),
                np_title_id=None,
            )
            self._pagination_args.increment_offset()
            yield title_trophy_sum

        offset = response.get("nextOffset") or 0
        if offset > 0:
            self._has_next = True
        else:
            self._has_next = False

    def get_trophy_summary_for_title(self) -> Generator[TrophyTitle, None, None]:
        """Retrieve a summary of the trophies earned by a user for specific titles.

        :param list[str] title_ids: Unique ID of the title

        :returns: Generator object with TitleTrophySummary objects

        :raises PSNAWPForbidden: If the user's profile is private

        """
        params = {"npTitleIds": ",".join(self.title_ids if self.title_ids is not None else [])}
        response = self.authenticator.get(url=self._url, params=params).json()
        self._total_item_count = response.get("totalItemCount", 0)

        for title in response.get("titles"):
            for trophy_title in title.get("trophyTitles"):
                title_trophy_sum = TrophyTitle(
                    np_service_name=trophy_title.get("npServiceName"),
                    np_communication_id=trophy_title.get("npCommunicationId"),
                    trophy_set_version=trophy_title.get("trophySetVersion"),
                    title_name=trophy_title.get("trophyTitleName"),
                    title_detail=trophy_title.get("trophyTitleDetail"),
                    title_icon_url=trophy_title.get("trophyTitleIconUrl"),
                    title_platform=frozenset(
                        [PlatformType(platform_str) for platform_str in trophy_title.get("trophyTitlePlatform", "").split(",")],
                    ),
                    has_trophy_groups=trophy_title.get("hasTrophyGroups"),
                    progress=trophy_title.get("progress"),
                    hidden_flag=trophy_title.get("hiddenFlag"),
                    last_updated_date_time=trophy_title.get("lastUpdatedDateTime"),
                    defined_trophies=trophy_title.get(
                        "definedTrophies",
                        {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                    ),
                    earned_trophies=trophy_title.get(
                        "earnedTrophies",
                        {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                    ),
                    np_title_id=title.get("npTitleId"),
                )
                yield title_trophy_sum

        # This endpoint does not have pagination
        self._has_next = False

    @staticmethod
    def get_np_communication_id(authenticator: Authenticator, title_id: str, account_id: str) -> str:
        """Returns the np communication id of title. This is required for requesting detail about a titles trophies.

        .. note::

            The endpoint only returns useful response back if the account has played that particular video game.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param title_id: Unique ID of the title
        :param account_id: Account ID of the user.

        :returns: np communication id of title

        :raises PSNAWPNotFound: If the user does not have any trophies for that game or the game doesn't exist.

        """
        params = {"npTitleIds": f"{title_id},"}

        try:
            response = authenticator.get(
                url=f"{BASE_PATH['trophies']}{API_PATH['trophy_titles_for_title'].format(account_id=account_id)}",
                params=params,
            ).json()
        except (PSNAWPBadRequest, PSNAWPNotFound) as bad_req:
            raise PSNAWPNotFound(f"Could not find a Video Game with Title: {title_id}") from bad_req

        if len(response.get("titles")[0].get("trophyTitles")) == 0:
            raise PSNAWPNotFound(f"Could not find a Video Game with Title: {title_id}. Most likely the user doesn't own the game.")

        np_comm_id: str = response.get("titles")[0].get("trophyTitles")[0].get("npCommunicationId", title_id)
        return np_comm_id



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generator, Literal, Optional, overload

from typing_extensions import Self

from psnawp_api.core import (
    PSNAWPBadRequest,
    PSNAWPForbidden,
    PSNAWPNotFound,
)
from psnawp_api.models.listing import PaginationArguments
from psnawp_api.models.title_stats import TitleStatsIterator
from psnawp_api.models.trophies import (
    TrophyGroupsSummaryBuilder,
    TrophyIterator,
    TrophySummary,
    TrophyTitleIterator,
    TrophyWithProgressIterator,
)
from psnawp_api.utils import API_PATH, BASE_PATH, extract_region_from_npid

if TYPE_CHECKING:
    from pycountry.db import Country

    from psnawp_api.core import Authenticator
    from psnawp_api.models.trophies import PlatformType, TrophyGroupsSummary, TrophyGroupSummary, TrophyGroupSummaryWithProgress


class User:
    """This class will contain the information about the PSN ID you passed in when creating object"""

    @classmethod
    def from_online_id(cls, authenticator: Authenticator, online_id: str) -> Self:
        """Creates the User instance from online ID and returns the instance.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param online_id: Online ID (GamerTag) of the user.

        :returns: User Class object which represents a PlayStation account

        :raises PSNAWPNotFound: If the user is not valid/found.

        """
        try:
            query = {"fields": "accountId,onlineId,currentOnlineId"}
            response: dict[str, Any] = authenticator.get(
                url=f"{BASE_PATH['legacy_profile_uri']}{API_PATH['legacy_profile'].format(online_id=online_id)}",
                params=query,
            ).json()
            account_id = response["profile"]["accountId"]
            online_id = response["profile"].get("currentOnlineId") or response["profile"].get("onlineId")
            return cls(authenticator, online_id, account_id)
        except PSNAWPNotFound as not_found:
            raise PSNAWPNotFound(f"Online ID {online_id} does not exist.") from not_found

    @classmethod
    def from_account_id(cls, authenticator: Authenticator, account_id: str) -> Self:
        """Creates the User instance from account ID and returns the instance.

        :param request_builder: Used to call http requests.
        :param account_id: Account ID of the user.

        :returns: User Class object which represents a PlayStation account

        :raises PSNAWPNotFound: If the user is not valid/found.

        """
        try:
            response: dict[str, Any] = authenticator.get(url=f"{BASE_PATH['profile_uri']}{API_PATH['profiles'].format(account_id=account_id)}").json()
            return cls(authenticator, response["onlineId"], account_id)
        except PSNAWPBadRequest as bad_request:
            raise PSNAWPNotFound(f"Account ID {account_id} does not exist.") from bad_request

    def __init__(
        self,
        authenticator: Authenticator,
        online_id: str,
        account_id: str,
    ):
        """Constructor of Class User.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param online_id: Online ID (GamerTag) of the user.
        :param account_id: Account ID of the user.

        .. note::

            This class is intended to be interfaced with through PSNAWP.

        """
        self.authenticator = authenticator
        self.online_id = online_id
        self.account_id = account_id
        self.prev_online_id = online_id

    def profile(self) -> dict[str, Any]:
        """Gets the profile of the user such as about me, avatars, languages etc...

        :returns: A dict containing info similar to what is shown below:

            .. literalinclude:: examples/user/profile.json
                :language: json


        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.profile())

        """

        response: dict[str, Any] = self.authenticator.get(url=f"{BASE_PATH['profile_uri']}{API_PATH['profiles'].format(account_id=self.account_id)}").json()
        return response

    def get_region(self) -> Optional[Country]:
        """Gets the region of the user.

        :returns: Returns Country object from Pycountry of the User or None if not found.

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.get_region())

        .. note::

            See https://github.com/pycountry/pycountry for more info on Country object.

        """
        response = self.get_profile_legacy()
        npid = response.get("profile", {}).get("npId", "")
        return extract_region_from_npid(npid)

    def get_profile_legacy(self) -> dict[str, Any]:
        """Gets the user profile info from legacy api endpoint. Useful for legacy console (PS3, PS4) presence.

        :returns: A dict containing info similar to what is shown below:

            .. literalinclude:: examples/client/get_profile_legacy.json
                :language: json


        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.get_profile_legacy())

        """
        params = {
            "fields": "npId,onlineId,accountId,avatarUrls,plus,aboutMe,languagesUsed,trophySummary(@default,level,progress,earnedTrophies),"
            "isOfficiallyVerified,personalDetail(@default,profilePictureUrls),personalDetailSharing,personalDetailSharingRequestMessageFlag,"
            "primaryOnlineStatus,presences(@default,@titleInfo,platform,lastOnlineDate,hasBroadcastData),requestMessageFlag,blocking,friendRelation,"
            "following,consoleAvailability"
        }

        response: dict[str, Any] = self.authenticator.get(
            url=f"{BASE_PATH['legacy_profile_uri']}{API_PATH['legacy_profile'].format(online_id=self.online_id)}",
            params=params,
        ).json()

        return response

    def get_presence(self) -> dict[str, Any]:
        """Gets the presences of a user. If the profile is private

        :returns: A dict containing info similar to what is shown below:

            .. literalinclude:: examples/user/get_presence.json
                :language: json


        :raises PSNAWPForbidden: When the user's profile is private, and you don't have permission to view their online status.

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.get_presence())

        """
        try:
            params = {"type": "primary"}
            response: dict[str, Any] = self.authenticator.get(
                url=f"{BASE_PATH['profile_uri']}/{self.account_id}{API_PATH['basic_presences']}",
                params=params,
            ).json()
            return response
        except PSNAWPForbidden as forbidden:
            raise PSNAWPForbidden(f"You are not allowed to check the presence of user {self.online_id}") from forbidden

    def friendship(self) -> dict[str, Any]:
        """Gets the friendship status and stats of the user

        :returns: A dict containing info similar to what is shown below

            .. literalinclude:: examples/user/friendship.json
                :language: json


        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.friendship())

        """
        response: dict[Any, Any] = self.authenticator.get(
            url=f"{BASE_PATH['profile_uri']}{API_PATH['friends_summary'].format(account_id=self.account_id)}"
        ).json()
        return response

    def accept_friend_request(self) -> None:
        """Accept the friend request by the User.

        :returns: None
        """
        self.authenticator.put(url=f"{BASE_PATH['profile_uri']}{API_PATH['manage_friendship'].format(account_id=self.account_id)}")

    def remove_friend(self) -> None:
        """Decline the friend request or unfriend the User.

        :returns: None
        """
        self.authenticator.delete(url=f"{BASE_PATH['profile_uri']}{API_PATH['manage_friendship'].format(account_id=self.account_id)}")

    def friends_list(self, limit: int = 1000) -> Generator[User, None, None]:
        """Gets the friends list and returns an iterator of User objects.

        :param limit: The number of items from input max is 1000.

        :returns: All friends in your friends list.

        :raises PSNAWPForbidden: When the user's when you don't have permission to view their friends list.

        .. code-block:: Python

            client = psnawp.me()
            friends_list = client.friends_list()

            for friend in friends_list:
                ...

        """
        limit = min(1000, limit)

        params = {"limit": limit}
        response = self.authenticator.get(url=f"{BASE_PATH['profile_uri']}{API_PATH['friends_list'].format(account_id=self.account_id)}", params=params).json()
        return (
            User.from_account_id(
                authenticator=self.authenticator,
                account_id=account_id,
            )
            for account_id in response["friends"]
        )

    def is_blocked(self) -> bool:
        """Checks if the user is blocked by you

        :returns: True if the user is blocked otherwise False

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.is_blocked())

        """
        response = self.authenticator.get(url=f"{BASE_PATH['profile_uri']}{API_PATH['blocked_users']}").json()
        return self.account_id in response["blockList"]

    def get_shareable_profile_link(self) -> dict[str, str]:
        """Gets the shareable link and QR code for the PlayStation profile.

        This method fetches the URL that can be used to easily share the user's PlayStation profile.
        Additionally, it provides a QR code image URL that corresponds to the shareable URL.

        :returns: A dict containing info similar to what is shown below:

            .. literalinclude:: examples/client/shareable_profile.json
                :language: json

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.get_shareable_profile_link())

        """
        response: dict[str, str] = self.authenticator.get(url=f"{BASE_PATH['cpss']}{API_PATH['share_profile'].format(account_id=self.account_id)}").json()
        return response

    def trophy_summary(self) -> TrophySummary:
        """Retrieve an overall summary of the number of trophies earned for a user broken down by

        - type
        - overall trophy level
        - progress towards the next level
        - current tier

        :returns: Trophy Summary Object containing all information

        :raises PSNAWPForbidden: If the user's profile is private

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            print(user_example.trophy_summary())

        """
        return TrophySummary.from_endpoint(authenticator=self.authenticator, account_id=self.account_id)

    def trophy_titles(self, limit: Optional[int] = None, offset: int = 0, page_size: int = 50) -> TrophyTitleIterator:
        """Retrieve all game titles associated with an account, and a summary of trophies earned from them.

        :param limit: Limit of titles returned, None means to return all trophy titles.
        :param page_size: The number of items to receive per api request.
        :param offset: Specifies the offset for paginator.

        :returns: Generator object with TrophyTitle objects.

        :raises PSNAWPForbidden: If the user's profile is private.

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            for trophy_title in user_example.trophy_titles(limit=None):
                print(trophy_title)

        """
        pg_args = PaginationArguments(total_limit=limit, offset=offset, page_size=page_size)
        return TrophyTitleIterator.from_endpoint(authenticator=self.authenticator, pagination_args=pg_args, account_id=self.account_id, title_ids=None)

    def trophy_titles_for_title(self, title_ids: list[str]) -> TrophyTitleIterator:
        """Retrieve a summary of the trophies earned by a user for specific titles.

        :param list[str] title_ids: Unique ID of the title.

        :returns: Generator object with TrophyTitle objects.

        :raises PSNAWPForbidden: If the user's profile is private.

        .. note::

            ``title_id`` can be obtained from https://andshrew.github.io/PlayStation-Titles/ or from :py:class:`psnawp_api.models.search.Search`

        .. code-block:: Python

            user_example = psnawp.user(online_id="VaultTec_Trading")
            for trophy_title in user_example.trophy_titles_for_title(title_ids=["CUSA00265_00"]):
                print(trophy_title)

        """
        pg_args = PaginationArguments(total_limit=None, offset=0, page_size=0)  # Not used
        return TrophyTitleIterator.from_endpoint(authenticator=self.authenticator, pagination_args=pg_args, account_id=self.account_id, title_ids=title_ids)

    @overload
    def trophies(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: Literal[False] = False,
        trophy_group_id: str = "default",
        limit: Optional[int] = None,
        offset: int = 0,
        page_size: int = 200,
    ) -> TrophyIterator: ...
    @overload
    def trophies(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: Literal[True],
        trophy_group_id: str = "default",
        limit: Optional[int] = None,
        offset: int = 0,
        page_size: int = 200,
    ) -> TrophyWithProgressIterator: ...
    def trophies(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: bool = False,
        trophy_group_id: str = "default",
        limit: Optional[int] = None,
        offset: int = 0,
        page_size: int = 200,
    ) -> TrophyIterator | TrophyWithProgressIterator:
        """Retrieves all trophies for a specified group within a game title, optionally including user progress.

        :param np_communication_id: Unique ID of a game title used to request trophy information. This can be obtained from ``GameTitle`` class.
        :param platform: The platform this title belongs to.
        :param trophy_group_id: ID for the trophy group. Each game expansion is represented by a separate ID. all to return all trophies for the title, default
            for the game itself, and additional groups starting from 001 and so on return expansions trophies.
        :param limit: Maximum number of trophies to return. None means all available trophies will be returned.
        :param include_progress: If True, includes progress information for each trophy.
        :param offset: The starting point within the collection of trophies.
        :param page_size: The number of trophies to return per page.

        :returns: Returns the Trophy Generator object with all the information

        :raises PSNAWPNotFound: If you don't have any trophies for that game.
        :raises PSNAWPForbidden: If the user's profile is private

        .. warning::

            Setting ``include_progress`` to ``True`` will consume more rate limits as progress information is fetched from a separate endpoint.

        """

        pg_args = PaginationArguments(total_limit=limit, offset=offset, page_size=page_size)
        if not include_progress:
            return TrophyIterator.from_endpoint(
                authenticator=self.authenticator,
                pagination_args=pg_args,
                np_communication_id=np_communication_id,
                platform=platform,
                trophy_group_id=trophy_group_id,
            )
        else:
            return TrophyWithProgressIterator.from_endpoint(
                authenticator=self.authenticator,
                pagination_args=pg_args,
                np_communication_id=np_communication_id,
                platform=platform,
                trophy_group_id=trophy_group_id,
                account_id=self.account_id,
            )

    @overload
    def trophy_groups_summary(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: Literal[False] = False,
    ) -> TrophyGroupsSummary[TrophyGroupSummary]: ...
    @overload
    def trophy_groups_summary(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: Literal[True],
    ) -> TrophyGroupsSummary[TrophyGroupSummaryWithProgress]: ...
    def trophy_groups_summary(
        self,
        np_communication_id: str,
        platform: PlatformType,
        include_progress: bool = False,
    ) -> TrophyGroupsSummary[TrophyGroupSummary] | TrophyGroupsSummary[TrophyGroupSummaryWithProgress]:
        """Retrieves the trophy groups for a title and their respective trophy count.

        This is most commonly seen in games which have expansions where additional trophies are added.

        :param np_communication_id: Unique ID of the title used to request trophy information
        :param platform: The platform this title belongs to.
        :param platform: The platform this title belongs to.
        :param include_progress: If True, will fetch results from another endpoint and include progress for trophy group such as name and detail

        .. warning::

            Setting ``include_progress`` to ``True`` will use twice the amount of rate limit since the API wrapper has to obtain progress from a separate
            endpoint.

        :returns: TrophyGroupSummary object containing title and title groups trophy information.

        :raises PSNAWPNotFound: If you don't have any trophies for that game.
        :raises PSNAWPForbidden: If the user's profile is private.

        """
        if not include_progress:
            return TrophyGroupsSummaryBuilder(
                authenticator=self.authenticator,
                np_communication_id=np_communication_id,
            ).game_title_trophy_groups_summary(platform=platform)
        else:
            return TrophyGroupsSummaryBuilder(
                authenticator=self.authenticator,
                np_communication_id=np_communication_id,
            ).earned_user_trophy_groups_summary(account_id=self.account_id, platform=platform)

    def title_stats(self, *, limit: Optional[int] = None, offset: int = 0, page_size: int = 200) -> TitleStatsIterator:
        """Retrieve a list of titles with their stats and basic meta-data

        :param limit: Limit of titles returned.
        :param page_size: The number of items to receive per api request.
        :param offset: Specifies the offset for paginator.

        .. warning::

            Only returns data for PS4 games and above.

        :returns: Iterator class for TitleStats

        .. code-block:: Python

            user_example = psnawp.user(online_id="jeranther")
            for title in user_example.title_stats():
                ...

        """
        pg_args = PaginationArguments(total_limit=limit, offset=offset, page_size=page_size)
        return TitleStatsIterator.from_endpoint(authenticator=self.authenticator, account_id=self.account_id, pagination_args=pg_args)

    def __repr__(self) -> str:
        return f"<User online_id:{self.online_id} account_id:{self.account_id}>"

    def __str__(self) -> str:
        return f"Online ID: {self.online_id} Account ID: {self.account_id}"



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from psnawp_api.core import PSNAWPForbidden
from psnawp_api.models.trophies.trophy_constants import TrophySet
from psnawp_api.utils import API_PATH, BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator


@dataclass(frozen=True)
class TrophySummary:
    """Class representing the overall summary of the number of trophies earned by a user."""

    account_id: str
    "The ID of the account being accessed"
    trophy_level: int
    "The overall trophy level"
    progress: int
    "Percentage process towards the next trophy level"
    tier: int
    "The tier this trophy level is in"
    earned_trophies: TrophySet
    "Number of trophies which have been earned by type"

    @classmethod
    def from_endpoint(cls, authenticator: Authenticator, account_id: str) -> TrophySummary:
        """Retrieve an overall summary of the number of trophies earned for a user broken down by

        - type
        - overall trophy level
        - progress towards the next level
        - current tier

        :returns: TrophySummary object with all the information
        :rtype: TrophySummary

        :raises PSNAWPForbidden: If the user's profile is private

        """
        try:
            response = authenticator.get(url=f"{BASE_PATH['trophies']}{API_PATH['trophy_summary'].format(account_id=account_id)}").json()
        except PSNAWPForbidden as forbidden:
            raise PSNAWPForbidden("The target user has set their trophies visibility to private.") from forbidden
        return cls(
            account_id=account_id,
            trophy_level=response.get("trophyLevel", -1),
            progress=response.get("progress", -1),
            tier=response.get("tier", -1),
            earned_trophies=TrophySet(
                **response.get(
                    "earnedTrophies",
                    {"bronze": 0, "silver": 0, "gold": 0, "platinum": 0},
                )
            ),
        )



from __future__ import annotations

import base64
import binascii
from datetime import datetime
from typing import Optional, cast

from pycountry import countries
from pycountry.db import Country


def iso_format_to_datetime(iso_format: Optional[str]) -> Optional[datetime]:
    return datetime.fromisoformat(iso_format.replace("Z", "+00:00")) if iso_format is not None else None


def extract_region_from_npid(npid: str) -> Optional[Country]:
    try:
        decoded_npid = base64.b64decode(npid).decode("utf-8")
    except (binascii.Error, UnicodeDecodeError):
        return None

    # Assuming a valid decoded npid format (e.g. VaultTec-Co@b7.us), extract the region (e.g. "US")
    if "@" in decoded_npid and "." in decoded_npid:
        region_candidate = decoded_npid.split(".")[-1]
        if len(region_candidate) == 2 and region_candidate.isalpha():
            return cast(Country, countries.get(alpha_2=region_candidate))

    return None



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generator, Generic, Optional, TypeVar

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator

T = TypeVar("T")


class PaginationIterator(Iterator[T], Generic[T]):
    """Iterator class for iterating over paginated API endpoints."""

    def __init__(self, *, authenticator: Authenticator, url: str, pagination_args: PaginationArguments):
        """Initialize the PaginationIterator.

        :param authenticator: An instance of Authenticator for making API requests.
        :param url: The URL of the endpoint.
        :param params: Dictionary of parameters to be passed in the API request.

        """
        self.authenticator = authenticator
        self._url = url
        self._pagination_args = pagination_args

        self.__iterator: Optional[Generator[T, None, None]] = None
        self._has_next = False
        self._total_item_count = 0

    def __iter__(self) -> PaginationIterator[T]:
        """Return the iterator object.

        :returns: The iterator object.

        """
        return self

    def __next__(self) -> T:
        if self.__iterator is None:
            self.__iterator = self.fetch_next_page()
        try:
            return self.__iterator.__next__()
        except StopIteration:  # If all items on single page have been yielded
            # If we run out of pages
            if not self._has_next:
                raise StopIteration

            # If limit is reached
            if self._pagination_args.total_limit is not None:
                if self._pagination_args.is_limit_reached():
                    raise StopIteration

            # If there are pages remaining and limit is not reached
            self.__iterator = self.fetch_next_page()
            return self.__next__()

    def __len__(self) -> int:
        return self._total_item_count

    @abstractmethod
    def fetch_next_page(self) -> Generator[T, None, None]:
        """Fetch the next page of items from the API.

        .. note::

            The implementation of this methods are also responsible for incrementing the offset using ``increment_offset()`` and setting the
            ``self._total_item_count``.

        """
        raise NotImplementedError("Subclasses must implement the fetch_next_page method")

    def set_offset(self, offset: int) -> None:
        """Set the offset parameter for the API request.

        :param offset: The offset value to set.

        """
        self._pagination_args.offset = offset

    def set_page_size(self, page_size: int) -> None:
        """Set the page size (limit) parameter for the API request.

        :param page_size: The page size value to set.

        """
        self._pagination_args.page_size = page_size


@dataclass
class PaginationArguments:
    total_limit: Optional[int]
    page_size: int
    offset: int

    def get_params_dict(self) -> dict[str, int]:
        return {"limit": self.adjusted_page_size, "offset": self.offset}

    def increment_offset(self) -> None:
        self.offset += 1

    def is_limit_reached(self) -> bool:
        return self.offset == self.total_limit

    @property
    def adjusted_page_size(self) -> int:
        """Calculates the adjusted page size based on the total limit.

        If total_limit is None, returns the original page_size. If total_limit is less than page_size, returns total_limit.

        """
        if self.total_limit is None:
            return self.page_size
        return min(self.page_size, self.total_limit - self.offset)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Optional, TypedDict

from typing_extensions import Self

from psnawp_api.core import (
    PSNAWPBadRequest,
    PSNAWPForbidden,
    PSNAWPNotFound,
)
from psnawp_api.utils import API_PATH, BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.user import User


class Sender(TypedDict):
    accountId: str
    onlineId: str


class LatestMessage(TypedDict):
    alternativeMessageType: int
    body: str
    createdTimestamp: str
    messageType: int
    messageUid: str
    sender: Sender


class MainThread(TypedDict):
    latestMessage: LatestMessage
    modifiedTimestamp: str
    readMessageUid: str
    threadId: str


class Member(TypedDict):
    accountId: str
    onlineId: str
    role: str


class GroupIcon(TypedDict):
    status: int


class GroupName(TypedDict):
    status: int
    value: str


class NotificationSetting(TypedDict):
    isMute: bool


class GroupDetails(TypedDict):
    existsNewArrival: bool
    groupIcon: GroupIcon
    groupId: str
    groupName: GroupName
    groupType: int
    isFavorite: bool
    joinedTimestamp: str
    mainThread: MainThread
    members: list[Member]
    modifiedTimestamp: str
    notificationSetting: NotificationSetting


class MessageResponse(TypedDict):
    messageUid: str
    createdTimestamp: str


class Group:
    """The Group class manages PSN group endpoints related to messages (Group and Direct Messages)."""

    def __init__(
        self,
        authenticator: Authenticator,
        group_id: Optional[str],
    ) -> None:
        """Constructor of Group.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param group_id: The Group ID of a group.

        :raises PSNAWPNotFound: If group id does not exist or is invalid.
        :raises PSNAWPForbidden: If you are sending message a user who has blocked you.

        .. note::

            This class is intended to be interfaced with through PSNAWP.

        """

        self.authenticator = authenticator
        self.group_id = group_id

    @classmethod
    def create_from_group_id(cls, authenticator: Authenticator, group_id: str) -> Self:
        return cls(authenticator, group_id)

    @classmethod
    def create_from_users(cls, authenticator: Authenticator, users: Iterable[User]) -> Self:
        """Creates a new group from the provide list of Users.

        :raises PSNAWPForbidden: If you are sending message a user who has blocked you.

        """
        account_ids = [{"accountId": user.account_id} for user in users]
        invitees = {"invitees": account_ids}

        try:
            response = authenticator.post(
                url=f"{BASE_PATH['gaming_lounge']}{API_PATH['create_group']}",
                json=invitees,
            ).json()
            return cls(authenticator, response["groupId"])
        except PSNAWPForbidden as forbidden:
            raise PSNAWPForbidden("Can't create group because of the players' privacy settings.") from forbidden

    def change_name(self, group_name: str) -> None:
        """Changes the group name to one specified in arguments.

        :param group_name: The name of the group that will be set.

        :returns: None

        :raises PSNAWPBadRequest: If you are not part of group or the group is a DM.

        .. note::

            You cannot change the name of DM groups. i.e. Groups with only two people (including you).

        """

        data = {"groupName": {"value": group_name}}
        try:
            self.authenticator.patch(
                url=f"{BASE_PATH['gaming_lounge']}{API_PATH['group_settings'].format(group_id=self.group_id)}",
                json=data,
            )
        except PSNAWPBadRequest as bad_req:
            raise PSNAWPBadRequest(f"The group name of Group ID {self.group_id} does cannot be changed. Group is either a dm or does not exist.") from bad_req

    def get_group_information(self) -> GroupDetails:
        """Gets the group chat information such as about me, avatars, languages etc...

        :returns: A dict containing info similar to what is shown below:

        :raises PSNAWPNotFound: If group id does not exist or is invalid.

        .. literalinclude:: examples/group/group_information.json
            :language: json

        """

        param = {
            "includeFields": "groupName,groupIcon,members,mainThread,joinedTimestamp,modifiedTimestamp,isFavorite,existsNewArrival,notificationSetting",
        }

        try:
            response: GroupDetails = self.authenticator.get(
                url=f"{BASE_PATH['gaming_lounge']}{API_PATH['group_members'].format(group_id=self.group_id)}",
                params=param,
            ).json()

            return response
        except PSNAWPNotFound as not_found:
            raise PSNAWPNotFound(f"Group ID {self.group_id} does not exist.") from not_found

    def send_message(self, message: str) -> MessageResponse:
        """Sends a message in the group.

        .. note::

            For now only text based messaging is supported.

        :param message: Message Body

        :returns: A dict containing info similar to what is shown below:

        .. code-block:: json

            {
                "messageUid": "1#425961448584099",
                "createdTimestamp": "1663911908531"
            }

        """

        data = {"messageType": 1, "body": message}

        response: MessageResponse = self.authenticator.post(
            url=f"{BASE_PATH['gaming_lounge']}{API_PATH['send_group_message'].format(group_id=self.group_id)}",
            json=data,
        ).json()

        return response

    def get_conversation(self, limit: int = 20) -> dict[str, Any]:
        """Gets the conversations in a group.

        :param limit: The number of conversations to receive.

        :returns: A dict containing info similar to what is shown below:

            .. code-block::

                {
                    "messages": [
                        {
                            "messageUid": "1#425961448584099",
                            "messageType": 1,
                            "alternativeMessageType": 1,
                            "body": "Hello World",
                            "createdTimestamp": "1663911908531",
                            "sender": {
                                "accountId": "8520698476712646544",
                                "onlineId": "VaultTec_Trading"
                            }
                        }
                    ],
                    "previous": "1#425961448584099",
                    "next": "1#425961448584099",
                    "reachedEndOfPage": false,
                    "messageCount": 1
                }

        """

        param = {"limit": limit}

        response: dict[str, Any] = self.authenticator.get(
            url=f"{BASE_PATH['gaming_lounge']}{API_PATH['conversation'].format(group_id=self.group_id)}",
            params=param,
        ).json()

        return response

    def leave_group(self) -> None:
        """Leave the current group

        :raises PSNAWPNotFound: If you are not part of the group.

        """

        self.authenticator.delete(url=f"{BASE_PATH['gaming_lounge']}{API_PATH['leave_group'].format(group_id=self.group_id)}")

    def invite_members(self, users: Iterable[User]) -> None:
        """
        Invite users to join the group.

        If all users in the invite list have blocked you, a PSNAWPForbidden
        exception is raised. Users who have blocked you are skipped, and
        the remaining users are invited.

        :param users: An iterable of User objects to be invited.
        :type users: Iterable[User]

        :raises PSNAWPForbidden: If all users in the invite list have blocked you.
        """
        account_ids = [{"accountId": user.account_id} for user in users]
        invitees = {"invitees": account_ids}

        try:
            self.authenticator.post(
                url=f"{BASE_PATH['gaming_lounge']}{API_PATH['create_group']}",
                json=invitees,
            ).json()
        except PSNAWPForbidden as forbidden:
            raise PSNAWPForbidden("Can't add users to group because of the players' privacy settings.") from forbidden

    def kick_member(self, user: User) -> None:
        """
        Remove a user from the group.

        :param user: The User object representing the member to be removed.

        :raises PSNAWPNotFound: If the user is not in the group.
        """
        try:
            self.authenticator.delete(url=f"{BASE_PATH['gaming_lounge']}{API_PATH['kick_member'].format(group_id=self.group_id, account_id=user.account_id)}")
        except PSNAWPNotFound as not_found:
            raise PSNAWPNotFound(f'User "{user.online_id}" is not a member of the group.') from not_found

    def __repr__(self) -> str:
        return f"<Group group_id:{self.group_id}>"

    def __str__(self) -> str:
        return f"Group ID: {self.group_id}"



BASE_PATH = {
    "base_uri": "https://ca.account.sony.com/api",
    "profile_uri": "https://m.np.playstation.com/api/userProfile/v1/internal/users",
    "account_uri": "https://dms.api.playstation.com/api",
    "legacy_profile_uri": "https://us-prof.np.community.playstation.net/userProfile/v1/users",
    "gaming_lounge": "https://m.np.playstation.com/api/gamingLoungeGroups/v1",
    "universal_search": "https://m.np.playstation.com/api/search/v1/universalSearch",
    "game_titles": "https://m.np.playstation.com/api/catalog/v2/titles",
    "trophies": "https://m.np.playstation.com/api/trophy/v1",
    "games_list": "https://m.np.playstation.com/api/gamelist/v2",
    "graph_ql": "https://m.np.playstation.com/api/graphql/v1/op",
    "cpss": "https://m.np.playstation.com/api/cpss",
}

API_PATH = {
    # Auth Endpoints
    "oauth_code": "/authz/v3/oauth/authorize",
    "access_token": "/authz/v3/oauth/token",
    # Client/User Endpoint
    "my_account": "/v1/devices/accounts/me",
    "profiles": "/{account_id}/profiles",
    "legacy_profile": "/{online_id}/profile2",
    "friends_list": "/{account_id}/friends",
    "friends_request": "/{account_id}/friends/receivedRequests",
    "manage_friendship": "/me/friends/{account_id}",
    "friends_summary": "/me/friends/{account_id}/summary",
    "available_to_play": "/me/friends/subscribing/availableToPlay",
    "blocked_users": "/me/blocks",
    "basic_presences": "/basicPresences",
    "share_profile": "/v1/share/profile/{account_id}",
    # Group Endpoints
    "my_groups": "/members/me/groups",
    "group_settings": "/groups/{group_id}",
    "create_group": "/groups",
    "group_members": "/members/me/groups/{group_id}",
    "invite_members": "/groups/{group_id}/invitees",
    "kick_member": "/groups/{group_id}/members/{account_id}",
    "send_group_message": "/groups/{group_id}/threads/{group_id}/messages",
    "conversation": "/members/me/groups/{group_id}/threads/{group_id}/messages",
    "leave_group": "/groups/{group_id}/members/me",
    # Game Title Details
    "title_concept": "/{title_id}/concepts",
    # Trophy Endpoints
    "trophy_titles": "/users/{account_id}/trophyTitles",
    "trophies_for_title": "/npCommunicationIds/{np_communication_id}/trophyGroups/{trophy_group_id}/trophies",
    "trophies_earned_for_title": "/users/{account_id}/npCommunicationIds/{np_communication_id}/trophyGroups/{trophy_group_id}/trophies",
    "trophy_summary": "/users/{account_id}/trophySummary",
    "title_trophy_group": "/npCommunicationIds/{np_communication_id}/trophyGroups",
    "user_title_trophy_group": "/users/{account_id}/npCommunicationIds/{np_communication_id}/trophyGroups",
    "trophy_titles_for_title": "/users/{account_id}/titles/trophyTitles",
    # User Title Data
    "user_game_data": "/users/{account_id}/titles",
}



from __future__ import annotations

import time
import uuid
from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional, TypedDict, TypeVar, cast
from urllib.parse import parse_qs, urlparse

from requests import Response
from typing_extensions import NotRequired, ParamSpec, Unpack

from psnawp_api.core.psnawp_exceptions import PSNAWPAuthenticationError
from psnawp_api.core.request_builder import RequestBuilder
from psnawp_api.utils import API_PATH, BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core.request_builder import RequestBuilderHeaders, RequestOptions

PT = ParamSpec("PT")
RT = TypeVar("RT")


def pre_request_processing(method: Callable[PT, RT]) -> Callable[PT, RT]:
    @wraps(method)
    def _impl(*method_args: PT.args, **method_kwargs: PT.kwargs) -> RT:
        authenticator_obj = cast(Authenticator, method_args[0])
        if authenticator_obj.token_response is None:
            authorization_code = authenticator_obj.get_authorization_code()
            authenticator_obj.fetch_access_token_from_authorization(authorization_code)
        else:
            authenticator_obj.fetch_access_token_from_refresh()

        method_out = method(*method_args, **method_kwargs)
        return method_out

    return _impl


class TokenResponse(TypedDict):
    access_token_expires_at: NotRequired[float]
    access_token: str
    expires_in: int
    id_token: str
    refresh_token_expires_at: NotRequired[float]
    refresh_token_expires_in: int
    refresh_token: str
    scope: str
    token_type: str


class Authenticator:
    """Provides an interface for PSN Authentication and API"""

    __CONSTANTS = {
        "CLIENT_ID": "09515159-7237-4370-9b40-3806e67c0891",
        "SCOPE": "psn:mobile.v2.core psn:clientapp",
        "REDIRECT_URI": "com.scee.psxandroid.scecompcall://redirect",
    }
    __AUTH_HEADER = {"Authorization": "Basic MDk1MTUxNTktNzIzNy00MzcwLTliNDAtMzgwNmU2N2MwODkxOnVjUGprYTV0bnRCMktxc1A="}

    def __init__(
        self,
        npsso_cookie: str,
        common_headers: RequestBuilderHeaders,
    ):
        """Represents a single authentication to PSN API.

        :param npsso_cookie: npsso cookie obtained from PSN website.
        :param common_headers: Common headers that will be added to all HTTP request.

        """
        self.npsso_token = npsso_cookie
        self.common_headers = common_headers
        self.request_builder = RequestBuilder(common_headers)
        self.token_response: Optional[TokenResponse] = None

        self.cid = str(uuid.UUID(int=uuid.getnode()))

    @property
    def access_token_expiration_time(self) -> float:
        """Get the access token expiration time.

        If the ``token_response`` is not available or ``access_token_expires_at``, returns current time.

        :returns: The expiration time of the access token as a Unix timestamp.

        """
        if self.token_response is None:
            return time.time()
        return self.token_response.get("access_token_expires_at", time.time())

    @property
    def refresh_token_expiration_time(self) -> float:
        """Get the refresh token expiration time.

        If the ``token_response`` is not available or ``refresh_token_expires_at``, returns current time.

        :returns: The expiration time of the refresh token as a Unix timestamp.

        """
        if self.token_response is None:
            return time.time()
        return self.token_response.get("refresh_token_expires_at", time.time())

    @property
    def access_token_expiration_in(self) -> int:
        """Get the time until the access token expires.

        If the ``token_response`` is not available or ``expires_in``, returns 0.

        :returns: The number of seconds until the access token expires.

        """
        if self.token_response is None:
            return 0
        return self.token_response.get("expires_in", 0)

    @property
    def refresh_token_expiration_in(self) -> int:
        """Get the time until the refresh token expires.

        If the ``token_response`` is not available or ``refresh_token_expires_in``, returns 0.

        :returns: The number of seconds until the refresh token expires.

        """
        if self.token_response is None:
            return 0
        return self.token_response.get("refresh_token_expires_in", 0)

    def fetch_access_token_from_refresh(self) -> None:
        """Updates the access token using refresh token."""

        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to obtain access_token using refresh token when refresh token is missing.")

        if self.access_token_expiration_time > time.time():
            return None

        header = type(self).__AUTH_HEADER | {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "com.sony.snei.np.android.sso.share.oauth.versa.USER_AGENT",
        }
        data = {
            "refresh_token": self.token_response["refresh_token"],
            "grant_type": "refresh_token",
            "scope": type(self).__CONSTANTS["SCOPE"],
            "token_format": "jwt",
        }
        response = self.request_builder.post(
            url=f"{BASE_PATH['base_uri']}{API_PATH['access_token']}",
            headers=header,
            data=data,
        )
        self.token_response = cast(TokenResponse, response.json())
        self.token_response["access_token_expires_at"] = self.token_response["expires_in"] + time.time()
        self.token_response["refresh_token_expires_at"] = self.token_response["refresh_token_expires_in"] + time.time()

    def fetch_access_token_from_authorization(self, authorization_code: str) -> None:
        """Obtain the access token using authorization code for the first time, after this the access token is obtained via refresh token.

        :param code: Code obtained using npsso code.

        """

        header = type(self).__AUTH_HEADER | {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "com.sony.snei.np.android.sso.share.oauth.versa.USER_AGENT",
            "X-Psn-Correlation-Id": self.cid,
        }
        data = {
            "cid": self.cid,
            "code": authorization_code,
            "grant_type": "authorization_code",
            "redirect_uri": type(self).__CONSTANTS["REDIRECT_URI"],
            "scope": type(self).__CONSTANTS["SCOPE"],
            "token_format": "jwt",
        }
        response = self.request_builder.post(
            url=f"{BASE_PATH['base_uri']}{API_PATH['access_token']}",
            headers=header,
            data=data,
        )
        self.token_response = cast(TokenResponse, response.json())
        self.token_response["access_token_expires_at"] = self.token_response["expires_in"] + time.time()
        self.token_response["refresh_token_expires_at"] = self.token_response["refresh_token_expires_in"] + time.time()

    def get_authorization_code(self) -> str:
        """Obtains the authorization code for PSN authentication.

        Obtains the access code and the refresh code. Access code lasts about 1 hour. While the refresh code lasts about 2 months. After 2 months a new npsso
        code is needed.

        :raises PSNAWPAuthenticationError: If authorization is not successful.

        """
        headers = {
            "Cookie": f"npsso={self.npsso_token}",
            "Content-Type": "application/x-www-form-urlencoded",
            "X-Requested-With": "com.scee.psxandroid",
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "same-site",
            "Sec-Fetch-User": "?1",
        }
        params = {
            "access_type": "offline",
            "cid": self.cid,
            "client_id": type(self).__CONSTANTS["CLIENT_ID"],
            "device_base_font_size": "10",
            "device_profile": "mobile",
            "elements_visibility": "no_aclink",
            "enable_scheme_error_code": "true",
            "no_captcha": "true",
            "PlatformPrivacyWs1": "minimal",
            "redirect_uri": type(self).__CONSTANTS["REDIRECT_URI"],
            "response_type": "code",
            "scope": type(self).__CONSTANTS["SCOPE"],
            "service_entity": "urn:service-entity:psn",
            "service_logo": "ps",
            "smcid": "psapp:signin",
            "support_scheme": "sneiprls",
            "turnOnTrustedBrowser": "true",
            "ui": "pr",
        }
        response = self.request_builder.get(
            url=f"{BASE_PATH['base_uri']}{API_PATH['oauth_code']}",
            headers=headers,
            params=params,
            allow_redirects=False,
        )

        location_url = response.headers["location"]
        parsed_url = urlparse(location_url)
        parsed_query = parse_qs(parsed_url.query)
        if "error" in parsed_query.keys():
            if "4165" in parsed_query["error_code"]:
                raise PSNAWPAuthenticationError("Your npsso code has expired or is incorrect. Please generate a new code!")
            else:
                raise PSNAWPAuthenticationError("Something went wrong while authenticating")

        return parsed_query["code"][0]

    @pre_request_processing
    def get(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Make a GET request with automatic Bearer token authorization.

        This method simplifies making GET requests by automatically adding the necessary Authorization header with a Bearer token. You can pass any additional
        arguments or keyword arguments, which will be forwarded to the underlying request builder's ``get`` method.

        :param kwargs: Additional arguments to be forwarded to the ``get`` method of the request builder.

        :returns: The response from the GET request.

        :raises PSNAWPAuthenticationError: If the ``token_response`` is ``None``, indicating that an attempt to make an HTTP request was made without an access
            token.
        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        .. note::

            The ``pre_request_processing`` decorator ensures that ``token_response`` is usually set correctly. The check for ``self.token_response is None`` is
            a safeguard in case of unexpected issues.

        """
        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to make HTTP Request without access_token.")

        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Bearer {self.token_response['access_token']}"

        return self.request_builder.get(**kwargs)

    @pre_request_processing
    def post(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Make a POST request with automatic Bearer token authorization.

        This method simplifies making POST requests by automatically adding the necessary Authorization header with a Bearer token. You can pass any additional
        arguments or keyword arguments, which will be forwarded to the underlying request builder's ``post`` method.

        :param kwargs: Additional arguments to be forwarded to the ``post`` method of the request builder.

        :returns: The response from the POST request.

        :raises PSNAWPAuthenticationError: If the ``token_response`` is ``None``, indicating that an attempt to make an HTTP request was made without an access
            token.
        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        .. note::

            The ``pre_request_processing`` decorator ensures that ``token_response`` is usually set correctly. The check for ``self.token_response is None`` is
            a safeguard in case of unexpected issues.

        """
        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to make HTTP Request without access_token.")

        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Bearer {self.token_response['access_token']}"

        return self.request_builder.post(**kwargs)

    @pre_request_processing
    def patch(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Make a PATCH request with automatic Bearer token authorization.

        This method simplifies making POST requests by automatically adding the necessary Authorization header with a Bearer token. You can pass any additional
        arguments or keyword arguments, which will be forwarded to the underlying request builder's ``patch`` method.

        :param kwargs: Additional arguments to be forwarded to the ``patch`` method of the request builder.

        :returns: The response from the POST request.

        :raises PSNAWPAuthenticationError: If the ``token_response`` is ``None``, indicating that an attempt to make an HTTP request was made without an access
            token.
        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        .. note::

            The ``pre_request_processing`` decorator ensures that ``token_response`` is usually set correctly. The check for ``self.token_response is None`` is
            a safeguard in case of unexpected issues.

        """
        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to make HTTP Request without access_token.")

        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Bearer {self.token_response['access_token']}"

        return self.request_builder.patch(**kwargs)

    @pre_request_processing
    def delete(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Make a DELETE request with automatic Bearer token authorization.

        This method simplifies making POST requests by automatically adding the necessary Authorization header with a Bearer token. You can pass any additional
        arguments or keyword arguments, which will be forwarded to the underlying request builder's ``delete`` method.

        :param kwargs: Additional arguments to be forwarded to the ``delete`` method of the request builder.

        :returns: The response from the POST request.

        :raises PSNAWPAuthenticationError: If the ``token_response`` is ``None``, indicating that an attempt to make an HTTP request was made without an access
            token.
        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        .. note::

            The ``pre_request_processing`` decorator ensures that ``token_response`` is usually set correctly. The check for ``self.token_response is None`` is
            a safeguard in case of unexpected issues.

        """
        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to make HTTP Request without access_token.")

        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Bearer {self.token_response['access_token']}"

        return self.request_builder.delete(**kwargs)

    @pre_request_processing
    def put(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Make a PUT request with automatic Bearer token authorization.

        This method simplifies making PUT requests by automatically adding the necessary Authorization header with a Bearer token. You can pass any additional
        arguments or keyword arguments, which will be forwarded to the underlying request builder's ``put`` method.

        :param kwargs: Additional arguments to be forwarded to the ``put`` method of the request builder.

        :returns: The response from the PUT request.

        :raises PSNAWPAuthenticationError: If the ``token_response`` is ``None``, indicating that an attempt to make an HTTP request was made without an access
            token.
        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        .. note::

            The ``pre_request_processing`` decorator ensures that ``token_response`` is usually set correctly. The check for ``self.token_response is None`` is
            a safeguard in case of unexpected issues.

        """
        if self.token_response is None:
            raise PSNAWPAuthenticationError("Attempt to make HTTP Request without access_token.")

        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Bearer {self.token_response['access_token']}"

        return self.request_builder.put(**kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from psnawp_api.models.listing import PaginationArguments
from psnawp_api.models.trophies import TrophyIterator
from psnawp_api.models.trophies.trophy_group import (
    TrophyGroupsSummary,
    TrophyGroupsSummaryBuilder,
)
from psnawp_api.models.trophies.trophy_titles import TrophyTitleIterator
from psnawp_api.utils import API_PATH, BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.trophies import PlatformType, TrophyGroupsSummary, TrophyGroupSummary


class GameTitle:
    """The GameTitle class provides the information and methods for retrieving Game details and trophies.

    .. note::

        This class is only useful if the user has played that video game. See :py:meth:`psnawp_api.psnawp.PSNAWP.game_title` for more information.

    """

    def __init__(self, authenticator: Authenticator, title_id: str, account_id: str, np_communication_id: Optional[str]):
        """The GameTitle class constructor.

        .. note::

            This class is intended to be interfaced with through PSNAWP.

        .. note::

            During the construction of the object, an additional call is made to get the np_communication_id. This ID is important for getting trophy data. You
            can avoid this additional HTTP request by passing in ``np_communication_id`` if you have it already.

        .. note::

            ``title_id`` can be obtained from https://andshrew.github.io/PlayStation-Titles/ or from :py:class:`psnawp_api.models.search.Search`

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param title_id: unique id of game.
        :param: account_id: The account whose trophy list is being accessed.
        :param np_communication_id: Unique ID of a game title used to request trophy information.

        :raises PSNAWPNotFound: If the user does not have any trophies for that game or the game doesn't exist.

        """
        self.authenticator = authenticator
        self.title_id = title_id
        if np_communication_id is None:
            self.np_communication_id = TrophyTitleIterator.get_np_communication_id(authenticator, title_id, account_id)
        else:
            self.np_communication_id = np_communication_id

    def get_details(self) -> list[dict[str, Any]]:
        """Get game details such as full name, description, genre, promotional videos/images, etc...

        :returns: A list of dicts containing info similar to what is shown below (Not all values are shown because of space limitations):

            .. literalinclude:: examples/game_title/get_details.json
                :language: json

        """

        param: dict[str, int | str] = {"age": 99, "country": "US", "language": "en-US"}

        response: list[dict[str, Any]] = self.authenticator.get(
            url=f"{BASE_PATH['game_titles']}{API_PATH['title_concept'].format(title_id=self.title_id)}",
            params=param,
        ).json()

        return response

    def trophies(
        self,
        platform: PlatformType,
        trophy_group_id: str = "default",
        limit: Optional[int] = None,
        offset: int = 0,
        page_size: int = 200,
    ) -> TrophyIterator:
        """Retrieves the individual trophy detail of a single - or all - trophy groups for a title.

        :param platform: The platform this title belongs to.
        :param trophy_group_id: ID for the trophy group. Each game expansion is represented by a separate ID. all to return all trophies for the title, default
            for the game itself, and additional groups starting from 001 and so on return expansions trophies.
        :param limit: Limit of trophies returned, None means to return all trophy titles.
        :param offset: The starting point within the collection of trophies.
        :param page_size: The number of trophies to return per page.

        :returns: Returns the Trophy Generator object with all the information

        :raises PSNAWPNotFound: If you don't have any trophies for that game.

        """

        pg_args = PaginationArguments(total_limit=limit, offset=offset, page_size=page_size)
        return TrophyIterator.from_endpoint(
            authenticator=self.authenticator,
            pagination_args=pg_args,
            np_communication_id=self.np_communication_id,
            platform=platform,
            trophy_group_id=trophy_group_id,
        )

    def trophy_groups_summary(self, platform: PlatformType) -> TrophyGroupsSummary[TrophyGroupSummary]:
        """Retrieves the trophy groups for a title and their respective trophy count.

        This is most commonly seen in games which have expansions where additional trophies are added.

        :param platform: The platform this title belongs to.

        :returns: TrophyGroupSummary object containing title and title groups trophy information.

        :raises PSNAWPNotFound: If you don't have any trophies for that game.

        """
        return TrophyGroupsSummaryBuilder(
            authenticator=self.authenticator,
            np_communication_id=self.np_communication_id,
        ).game_title_trophy_groups_summary(platform=platform)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum


class TrophyType(Enum):
    BRONZE = "bronze"
    SILVER = "silver"
    GOLD = "gold"
    PLATINUM = "platinum"


class PlatformType(Enum):
    UNKNOWN = "UNKNOWN"
    PS_VITA = "PSVITA"
    PS3 = "PS3"
    PS4 = "PS4"
    PS5 = "PS5"
    PSPC = "PSPC"

    @classmethod
    def _missing_(cls, value: object) -> "PlatformType":
        _ = value
        return cls.UNKNOWN


class TrophyRarity(Enum):
    ULTRA_RARE = 0
    VERY_RARE = 1
    RARE = 2
    COMMON = 3


@dataclass(frozen=True)
class TrophySet:
    bronze: int = field(default=0)
    silver: int = field(default=0)
    gold: int = field(default=0)
    platinum: int = field(default=0)



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any, TypedDict, cast

from pyrate_limiter import Duration, Limiter, RequestRate, SQLiteBucket
from requests import Response
from requests_ratelimiter import LimiterSession
from typing_extensions import NotRequired, TypeAlias, Unpack

from psnawp_api.core.psnawp_exceptions import (
    PSNAWPBadRequest,
    PSNAWPClientError,
    PSNAWPForbidden,
    PSNAWPNotAllowed,
    PSNAWPNotFound,
    PSNAWPServerError,
    PSNAWPTooManyRequests,
    PSNAWPUnauthorized,
)

if TYPE_CHECKING:
    from requests.sessions import RequestsCookieJar, _Auth, _Cert, _Data, _Files, _HooksInput, _Params, _Timeout, _Verify

request_builder_logger = getLogger("psnawp")


def response_checker(response: Response) -> None:
    """Checks the HTTP(S) response and raises corresponding PSNAWP exceptions.

    This function examines the status code of the HTTP response and raises PSNAWP exceptions based on error conditions.

    :param requests.Response response: The HTTP response object.

    :raises PSNAWPBadRequest: If the HTTP response status code is 400.
    :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
    :raises PSNAWPForbidden: If the HTTP response status code is 403.
    :raises PSNAWPNotFound: If the HTTP response status code is 404.
    :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
    :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
    :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
    :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

    """
    if response.status_code == 400:
        raise PSNAWPBadRequest(response.text)
    elif response.status_code == 401:
        raise PSNAWPUnauthorized(response.text)
    elif response.status_code == 403:
        raise PSNAWPForbidden(response.text)
    elif response.status_code == 404:
        raise PSNAWPNotFound(response.text)
    elif response.status_code == 405:
        raise PSNAWPNotAllowed(response.text)
    elif response.status_code == 429:
        raise PSNAWPTooManyRequests(response.text)
    elif 400 <= response.status_code < 500:
        raise PSNAWPClientError(response.text)
    elif response.status_code >= 500:
        raise PSNAWPServerError(response.text)
    else:
        response.raise_for_status()


RequestBuilderHeaders = TypedDict(
    "RequestBuilderHeaders",
    {
        "User-Agent": str,
        "Accept-Language": str,
        "Country": str,
    },
)

_TextMapping: TypeAlias = dict[str, str]


class RequestOptions(TypedDict):
    allow_redirects: NotRequired[bool]
    auth: NotRequired[_Auth]
    cert: NotRequired[_Cert]
    cookies: NotRequired[RequestsCookieJar | _TextMapping]
    data: NotRequired[_Data]
    files: NotRequired[_Files]
    headers: NotRequired[_TextMapping]
    hooks: NotRequired[_HooksInput]
    json: NotRequired[Any]
    params: NotRequired[_Params]
    proxies: NotRequired[_TextMapping]
    stream: NotRequired[bool]
    timeout: NotRequired[_Timeout]
    url: str | bytes
    verify: NotRequired[_Verify]


class RequestBuilder:
    """Handles all the HTTP Requests and provides a gateway to interacting with PSN API.

    :param common_headers: Default headers for the requests.

    """

    def __init__(self, common_headers: RequestBuilderHeaders) -> None:
        """Initialize Request Handler with default headers."""
        self.common_headers = cast(dict[str, str], common_headers)

        psn_api_rate = RequestRate(limit=300, interval=Duration.MINUTE * 15)
        limiter = Limiter(psn_api_rate, bucket_class=SQLiteBucket)
        self.session = LimiterSession(limiter=limiter, per_host=False, limit_statuses=[], burst=0)
        self.session.headers.update(self.common_headers)

    def request(self, method: str | bytes, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles HTTP requests and returns the requests.Response object.

        :param str | bytes method: The HTTP method to use for the request (e.g., GET, PATCH).
        :param Unpack[RequestOptions] kwargs: The options for the HTTP request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        request_builder_logger.debug(
            "Sending request: method=%s, url=%s, headers=%s, body=%s", method, kwargs.get("url"), kwargs.get("headers"), kwargs.get("data")
        )
        response = self.session.request(method=method, **kwargs)
        request_builder_logger.debug("Received response: status_code=%d, headers=%s, body=%s", response.status_code, response.headers, response.text)
        response_checker(response)
        return response

    def get(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the GET requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the GET request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="get", **kwargs)

    def patch(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the PATCH requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the PATCH request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="patch", **kwargs)

    def post(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the POST requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the POST request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="post", **kwargs)

    def put(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the PUT requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the PUT request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="put", **kwargs)

    def delete(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the DELETE requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the DELETE request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="delete", **kwargs)

    def head(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the HEAD requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the HEAD request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="head", **kwargs)

    def options(self, **kwargs: Unpack[RequestOptions]) -> Response:
        """Handles the OPTIONS requests and returns the requests.Response object.

        :param Unpack[RequestOptions] kwargs: The options for the OPTIONS request.

        :returns: The Request Response Object.
        :rtype: requests.Response

        :raises PSNAWPBadRequest: If the HTTP response status code is 400.
        :raises PSNAWPUnauthorized: If the HTTP response status code is 401.
        :raises PSNAWPForbidden: If the HTTP response status code is 403.
        :raises PSNAWPNotFound: If the HTTP response status code is 404.
        :raises PSNAWPNotAllowed: If the HTTP response status code is 405.
        :raises PSNAWPTooManyRequests: If the HTTP response status code is 429.
        :raises PSNAWPClientError: If the HTTP response status code is in the 4xx range (excluding those listed above).
        :raises PSNAWPServerError: If the HTTP response status code is 500 or above.

        """
        return self.request(method="options", **kwargs)



from __future__ import annotations

import json
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Generator, Optional

from typing_extensions import List, NotRequired, Self, TypedDict

from psnawp_api.models.listing import PaginationIterator
from psnawp_api.utils import BASE_PATH

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.listing import PaginationArguments


class Data(TypedDict):
    universalDomainSearch: NotRequired[UniversalDomainSearch]
    universalContextSearch: NotRequired[UniversalContextSearch]


class UniversalContextSearch(TypedDict):
    __typename: str
    results: List[Result]


class UniversalDomainSearch(TypedDict):
    __typename: str
    domain: str
    domainTitle: str
    next: str
    searchResults: List[SearchResult]
    totalResultCount: int
    zeroState: bool


class Result(TypedDict):
    __typename: str
    domain: str
    domainTitle: str
    next: str
    searchResults: List[SearchResult]
    totalResultCount: int
    zeroState: bool


class SearchResult(TypedDict):
    __typename: str
    highlight: Highlight
    id: str
    result: DefaultProductOrResult


class Highlight(TypedDict):
    __typename: str
    name: Optional[List[str]]


class DefaultProductOrResult(TypedDict):
    __typename: str
    id: str
    invariantName: str
    itemType: str
    localizedStoreDisplayClassification: Optional[str]
    media: List[Media]
    name: str
    platforms: List[str]
    price: Price
    storeDisplayClassification: Optional[None]
    type: NotRequired[str]
    defaultProduct: NotRequired[DefaultProductOrResult]


class Media(TypedDict):
    __typename: str
    role: str
    type: str
    url: str


class Price(TypedDict):
    __typename: str
    basePrice: str
    basePriceValue: int
    campaignId: Optional[Any]
    currencyCode: str
    discountText: Optional[Any]
    discountedPrice: str
    discountedValue: int
    endTime: Optional[Any]
    isExclusive: bool
    isFree: bool
    isTiedToSubscription: bool
    qualifications: Optional[List[Qualification]]
    rewardId: str
    serviceBranding: List[str]
    skuId: str
    upsellServiceBranding: List[str]
    upsellText: Optional[Any]


class Qualification(TypedDict):
    __typename: str
    type: str
    value: str


class SearchDomain(IntEnum):
    FULL_GAMES = 0
    ADD_ONS = 1


class UniversalSearch:
    def __init__(self, authenticator: Authenticator, pagination_args: PaginationArguments, search_query: str) -> None:
        self.authenticator = authenticator
        self.pagination_args = pagination_args
        self.search_query = search_query
        self._search_common_header = {
            "Content-Type": "Application/json",
            "apollographql-client-name": "PlayStationApp-Android",
            "apollographql-client-version": "24.4.1",
        }

        variables: dict[str, str | int] = {
            "searchTerm": search_query,
            "searchContext": "MobileUniversalSearchGame",
            "displayTitleLocale": "en-US",
        }

        extensions = {
            "persistedQuery": {
                "version": 1,
                "sha256Hash": "a2fbc15433b37ca7bfcd7112f741735e13268f5e9ebd5ffce51b85acc126f41d",
            }
        }

        self.params = {
            "operationName": "metGetContextSearchResults",
            "variables": json.dumps(variables),
            "extensions": json.dumps(extensions),
        }

    def search_full_game(self) -> Generator[SearchResult, None, None]:
        response: dict[str, Any] = self.authenticator.get(
            url=BASE_PATH["graph_ql"],
            headers=self._search_common_header,
            params=self.params,
        ).json()

        data: Data = response.get("data", {"universalDomainSearch": {"results": []}})
        universal_search: UniversalContextSearch | dict[str, list[Any]] = data.get("universalContextSearch", {"results": []})
        full_games = universal_search.get("results", [])[SearchDomain.FULL_GAMES]
        search_results = full_games["searchResults"]

        for search_result in search_results:
            if self.pagination_args.is_limit_reached():
                return

            self.pagination_args.increment_offset()
            yield search_result

        search_iter = UniversalDomainSearchIterator.from_endpoint(
            authenticator=self.authenticator,
            pagination_args=self.pagination_args,
            search_query=self.search_query,
            search_domain=SearchDomain.FULL_GAMES,
            next_cursor=full_games["next"],
        )

        for search_result in search_iter:
            yield search_result

    def search_add_onns(self) -> Generator[SearchResult, None, None]:
        response: dict[str, Any] = self.authenticator.get(
            url=BASE_PATH["graph_ql"],
            headers=self._search_common_header,
            params=self.params,
        ).json()

        data: Data = response.get("data", {"universalDomainSearch": {"results": []}})
        universal_search: UniversalContextSearch | dict[str, list[Any]] = data.get("universalContextSearch", {"results": []})
        addons = universal_search.get("results", [])[SearchDomain.ADD_ONS]
        search_results = addons["searchResults"]

        for search_result in search_results:
            if self.pagination_args.is_limit_reached():
                return

            self.pagination_args.increment_offset()
            yield search_result

        search_iter = UniversalDomainSearchIterator.from_endpoint(
            authenticator=self.authenticator,
            pagination_args=self.pagination_args,
            search_query=self.search_query,
            search_domain=SearchDomain.FULL_GAMES,
            next_cursor=addons["next"],
        )

        for search_result in search_iter:
            yield search_result


class UniversalDomainSearchIterator(PaginationIterator[SearchResult]):
    def __init__(
        self,
        authenticator: Authenticator,
        url: str,
        pagination_args: PaginationArguments,
        search_query: str,
        search_domain: SearchDomain,
        next_cursor: str,
    ) -> None:
        """The SearchGames class provides the information and methods for searching full games playstation network.

        .. note::

            This class is intended to be interfaced with through PSNAWP.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.

        """
        super().__init__(authenticator=authenticator, url=url, pagination_args=pagination_args)
        self.search_query = search_query
        self.search_domain = search_domain
        self._search_common_header = {
            "Content-Type": "Application/json",
            "apollographql-client-name": "PlayStationApp-Android",
            "apollographql-client-version": "24.4.1",
        }
        self.next_cursor = next_cursor

    @classmethod
    def from_endpoint(
        cls,
        authenticator: Authenticator,
        pagination_args: PaginationArguments,
        search_query: str,
        search_domain: SearchDomain,
        next_cursor: str,
    ) -> Self:
        return cls(
            authenticator=authenticator,
            url=BASE_PATH["graph_ql"],
            pagination_args=pagination_args,
            search_query=search_query,
            search_domain=search_domain,
            next_cursor=next_cursor,
        )

    def fetch_next_page(self) -> Generator[SearchResult, None, None]:
        """Fetches the next page of Search Result objects from the API.

        :yield: A generator yielding Result objects.

        """

        variables: dict[str, str | int] = {
            "searchTerm": self.search_query,
            "searchDomain": "MobileGames" if self.search_domain == SearchDomain.FULL_GAMES else "MobileAddOns",
            "pageSize": self._pagination_args.adjusted_page_size,
            "pageOffset": self._pagination_args.offset,
            "nextCursor": self.next_cursor,
        }

        extensions = {
            "persistedQuery": {
                "version": 1,
                "sha256Hash": "b51624299bd17b3799f77c9f097cc8887a04d3873f0329095976a841595bc902",
            }
        }

        params = {
            "operationName": "metGetDomainSearchResults",
            "variables": json.dumps(variables),
            "extensions": json.dumps(extensions),
        }

        response: dict[str, Any] = self.authenticator.get(
            url=BASE_PATH["graph_ql"],
            headers=self._search_common_header,
            params=params,
        ).json()

        default_instance: UniversalDomainSearch = {
            "__typename": "",
            "domain": "",
            "domainTitle": "",
            "next": "",
            "searchResults": [],
            "totalResultCount": 0,
            "zeroState": False,
        }

        data: Data = response.get("data", {"universalDomainSearch": default_instance})
        universal_search = data.get("universalDomainSearch", default_instance)

        search_results = universal_search["searchResults"]
        self._total_item_count = universal_search.get("totalResultCount", 0)
        self.next_cursor = universal_search.get("next", "")

        for search_result in search_results:
            if self._pagination_args.is_limit_reached():
                return

            self._pagination_args.increment_offset()
            yield search_result

        if self.next_cursor:
            self._has_next = True
        else:
            self._has_next = False



class PSNAWPException(Exception):
    """Base Exception for all PSNAWP Exceptions."""


class PSNAWPServerError(PSNAWPException):
    """Exception raised if there is a problem at the server."""


class PSNAWPClientError(PSNAWPException):
    """Exception raised if there is a problem at the client."""


class PSNAWPAuthenticationError(PSNAWPClientError):
    """Exception for authentication related errors."""


class PSNAWPBadRequest(PSNAWPClientError):
    """Exception raised if bad request is made to the endpoint."""


class PSNAWPIllegalArgumentError(PSNAWPClientError):
    """Exception raised if user gave wrong input to a function."""


class PSNAWPUnauthorized(PSNAWPClientError):
    """Exception for accessing an action is not allowed due to missing the right authorization."""


class PSNAWPForbidden(PSNAWPClientError):
    """Exception for accessing an action is not allowed due to insufficient rights to a resource."""


class PSNAWPNotFound(PSNAWPClientError):
    """Exception raised if resource not found."""


class PSNAWPNotAllowed(PSNAWPClientError):
    """Exception raised if resource doesn't support this method."""


class PSNAWPTooManyRequests(PSNAWPClientError):
    """Exception raised if client sends too many requests."""



from psnawp_api.models.client import Client
from psnawp_api.models.game_title import GameTitle
from psnawp_api.models.group import Group
from psnawp_api.models.search import SearchDomain, UniversalSearch
from psnawp_api.models.title_stats import TitleStats, TitleStatsIterator
from psnawp_api.models.user import User

__all__ = [
    "Client",
    "GameTitle",
    "Group",
    "SearchDomain",
    "UniversalSearch",
    "TitleStats",
    "TitleStatsIterator",
    "User",
]



from __future__ import annotations

import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Any, Generator, Literal, Optional

from typing_extensions import Self

from psnawp_api.models.listing import PaginationIterator
from psnawp_api.utils.endpoints import API_PATH, BASE_PATH
from psnawp_api.utils.misc import iso_format_to_datetime

if TYPE_CHECKING:
    from psnawp_api.core import Authenticator
    from psnawp_api.models.listing import PaginationArguments


class PlatformCategory(Enum):
    UNKNOWN = "unknown"
    PS4 = "ps4_game"
    PS5 = "ps5_native_game"

    @classmethod
    def _missing_(cls, value: object) -> Literal[PlatformCategory.UNKNOWN]:
        _ = value
        return cls.UNKNOWN


PT_REGEX = re.compile(r"PT(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?")


def play_duration_to_timedelta(play_duration: Optional[str]) -> timedelta:
    """Provides a timedelta object for the play duration PSN sends. If for some reason the string is malformed or None, timedelta will return 0

    Valid patters: PT243H18M48S, PT21M18S, PT18H, PT18H20S, PT4H21M

    :param play_duration: String from API

    :returns: String parsed into a timedelta object

    .. note::

        PSN API returns the duration in this format: PT243H18M48S. The maximum time Unit is Hours, it does not extend to Days or Months.

    """
    hours = 0
    minutes = 0
    seconds = 0

    if play_duration:
        match = re.search(PT_REGEX, play_duration)
        if not match:
            return timedelta(hours=hours, minutes=minutes, seconds=seconds)

        hours = int(match.group(1)) if match.group(1) else 0
        minutes = int(match.group(2)) if match.group(2) else 0
        seconds = int(match.group(3)) if match.group(3) else 0

    return timedelta(hours=hours, minutes=minutes, seconds=seconds)


@dataclass(frozen=True)
class TitleStats:
    """A class that represents a PlayStation Video Game Play Time Stats.

    .. note::

        This class is intended to be interfaced with through PSNAWP.

    """

    title_id: Optional[str]
    "Game title id"
    name: Optional[str]
    "Game title name"
    image_url: Optional[str]
    "Image URL"
    category: Optional[PlatformCategory]
    "Category/Platform Type"
    play_count: Optional[int]
    "Number of times the game has been played"
    first_played_date_time: Optional[datetime]
    "First time the game was played"
    last_played_date_time: Optional[datetime]
    "Last time the game was played"
    play_duration: Optional[timedelta]
    "Total time the game has been played. Example: PT1H51M21S"

    @classmethod
    def from_dict(cls, game_stats_dict: dict[str, Any]) -> TitleStats:
        title_instance = cls(
            title_id=game_stats_dict.get("titleId"),
            name=game_stats_dict.get("name"),
            image_url=game_stats_dict.get("imageUrl"),
            category=PlatformCategory(game_stats_dict.get("category")),
            play_count=game_stats_dict.get("playCount"),
            first_played_date_time=iso_format_to_datetime(game_stats_dict.get("firstPlayedDateTime")),
            last_played_date_time=iso_format_to_datetime(game_stats_dict.get("lastPlayedDateTime")),
            play_duration=play_duration_to_timedelta(game_stats_dict.get("playDuration")),
        )
        return title_instance


class TitleStatsIterator(PaginationIterator[TitleStats]):
    """An iterator for fetching and paginating through TitleStats objects from the PlayStation Network API."""

    def __init__(self, authenticator: Authenticator, url: str, pagination_args: PaginationArguments) -> None:
        super().__init__(authenticator=authenticator, url=url, pagination_args=pagination_args)

    def fetch_next_page(self) -> Generator[TitleStats, None, None]:
        """Fetches the next page of TitleStats objects from the API.

        :yield: A generator yielding TitleStats objects.

        """
        response = self.authenticator.get(url=self._url, params=self._pagination_args.get_params_dict()).json()
        self._total_item_count = response.get("totalItemCount", 0)

        titles: list[dict[str, Any]] = response.get("titles")
        for title in titles:
            title_instance = TitleStats.from_dict(title)
            self._pagination_args.increment_offset()
            yield title_instance

        offset = response.get("nextOffset") or 0
        if offset > 0:
            self._has_next = True
        else:
            self._has_next = False

    @classmethod
    def from_endpoint(cls, authenticator: Authenticator, account_id: str, pagination_args: PaginationArguments) -> Self:
        """Creates an instance of TitleStatsIterator from the given endpoint.

        :param authenticator: The Authenticator instance used for making authenticated requests to the API.
        :param account_id: The account ID for which to fetch title stats.
        :param pagination_args: Arguments for handling pagination, including limit, offset, and page size.

        :returns: An instance of TitleStatsIterator.

        """
        url = f"{BASE_PATH['games_list']}{API_PATH['user_game_data'].format(account_id=account_id)}"
        return cls(authenticator=authenticator, url=url, pagination_args=pagination_args)



from psnawp_api.core.authenticator import Authenticator
from psnawp_api.core.psnawp_exceptions import (
    PSNAWPAuthenticationError,
    PSNAWPBadRequest,
    PSNAWPClientError,
    PSNAWPException,
    PSNAWPForbidden,
    PSNAWPIllegalArgumentError,
    PSNAWPNotAllowed,
    PSNAWPNotFound,
    PSNAWPServerError,
    PSNAWPTooManyRequests,
    PSNAWPUnauthorized,
)
from psnawp_api.core.request_builder import RequestBuilder, RequestBuilderHeaders, RequestOptions

__all__ = [
    "Authenticator",
    "PSNAWPAuthenticationError",
    "PSNAWPBadRequest",
    "PSNAWPClientError",
    "PSNAWPException",
    "PSNAWPForbidden",
    "PSNAWPIllegalArgumentError",
    "PSNAWPNotAllowed",
    "PSNAWPNotFound",
    "PSNAWPServerError",
    "PSNAWPTooManyRequests",
    "PSNAWPUnauthorized",
    "RequestBuilder",
    "RequestBuilderHeaders",
    "RequestOptions",
]
